import collections
import ipaddress

from visidata import VisiData, vd, Sheet, options, Column, asyncthread, Progress, TsvSheet, getattrdeep, ColumnAttr, date, vlen, filesize

vd.option('pcap_internet', 'n', '(y/s/n) if save_dot includes all internet hosts separately (y), combined (s), or does not include the internet (n)')

protocols = collections.defaultdict(dict)  # ['ethernet'] = {[6] -> 'IP'}
_flags = collections.defaultdict(dict)  # ['tcp'] = {[4] -> 'FIN'}


url_oui = 'https://visidata.org/plugins/pcap/wireshark-oui.tsv'
url_iana = 'https://visidata.org/plugins/pcap/iana-ports.tsv'

oui = {}  # [macprefix (like '01:02:dd:0')] -> 'manufacturer'
services = {}  # [('tcp', 25)] -> 'smtp'

@VisiData.api
def open_pcap(vd, p):
    return PcapSheet(p.base_stem, source=p)

open_cap = open_pcap
open_pcapng = open_pcap
open_ntar = open_pcap

def manuf(mac):
    return oui.get(mac[:13]) or oui.get(mac[:10]) or oui.get(mac[:8])

def macaddr(addrbytes):
    mac = ':'.join('%02x' % b for b in addrbytes)
    return mac

def macmanuf(mac):
    manuf = oui.get(mac[:13])
    if manuf:
        return manuf + mac[13:]

    manuf = oui.get(mac[:10])
    if manuf:
        return manuf + mac[10:]

    manuf = oui.get(mac[:8])
    if manuf:
        return manuf + mac[8:]

    return mac


def norm_host(host):
    if not host:
        return None
    srcmac = str(host.macaddr)
    if srcmac == 'ff:ff:ff:ff:ff:ff': return None

    srcip = str(host.ipaddr)
    if srcip == '0.0.0.0' or srcip == '::': return None
    if srcip == '255.255.255.255': return None

    if host.ipaddr:
        if host.ipaddr.is_global:
            opt = options.pcap_internet
            if opt == 'n':
                return None
            elif opt == 's':
                return "internet"

        if host.ipaddr.is_multicast:
            # include in multicast  (minus dns?)
            return 'multicast'

    names = [host.hostname, host.ipaddr, macmanuf(host.macaddr)]
    return '\\n'.join(str(x) for x in names if x)


def FlagGetter(flagfield):
    def flags_func(fl):
        return ' '.join([flagname for f, flagname in _flags[flagfield].items() if fl & f])
    return flags_func


def init_pcap():
    if protocols:  # already init'ed
        return

    global dpkt, dnslib
    dpkt = vd.importExternal('dpkt')
    dnslib = vd.importExternal('dnslib')

    load_consts(protocols['ethernet'], dpkt.ethernet, 'ETH_TYPE_')
    load_consts(protocols['ip'], dpkt.ip, 'IP_PROTO_')
    load_consts(_flags['ip_tos'], dpkt.ip, 'IP_TOS_')
    load_consts(protocols['icmp'], dpkt.icmp, 'ICMP_')
    load_consts(_flags['tcp'], dpkt.tcp, 'TH_')

    load_oui(url_oui)
    load_iana(url_iana)


def read_pcap(f):

    try:
        return dpkt.pcapng.Reader(f.open_bytes())
    except ValueError:
        return dpkt.pcap.Reader(f.open_bytes())


@asyncthread
def load_oui(url):
    vsoui = TsvSheet('vsoui', source=vd.urlcache(url, days=30))
    vsoui.reload.__wrapped__(vsoui)
    for r in vsoui.rows:
        if r.prefix.endswith('/36'): prefix = r.prefix[:13]
        elif r.prefix.endswith('/28'): prefix = r.prefix[:10]
        else: prefix = r.prefix[:8]
        try:
            oui[prefix.lower()] = r.shortname
        except Exception as e:
            vd.exceptionCaught(e)


@asyncthread
def load_iana(url):
    ports_tsv = TsvSheet('ports_tsv', source=vd.urlcache(url, days=30))
    ports_tsv.reload.__wrapped__(ports_tsv)
    for r in ports_tsv.rows:
        try:
            services[(r.transport, int(r.port))] = r.service
        except Exception as e:
            vd.exceptionCaught(e)


class Host:
    dns = {}  # [ipstr] -> dnsname
    hosts = {}  # [macaddr] -> { [ipaddr] -> Host }

    @classmethod
    def get_host(cls, pkt, field='src'):
        mac = macaddr(getattr(pkt, field))
        machosts = cls.hosts.get(mac, None)
        if not machosts:
            machosts = cls.hosts[mac] = {}

        ipraw = getattrdeep(pkt, 'ip.'+field, None)
        if ipraw is not None:
            ip = ipaddress.ip_address(ipraw)
            if ip not in machosts:
                machosts[ip] = Host(mac, ip)
            return machosts[ip]
        else:
            if machosts:
                return list(machosts.values())[0]

        return Host(mac, None)

    @classmethod
    def get_by_ip(cls, ip):
        'Returns Host instance for the given ip address.'
        ret = cls.hosts_by_ip.get(ip)
        if ret is None:
            ret = cls.hosts_by_ip[ip] = [Host(ip)]
        return ret

    def __init__(self, mac, ip):
        self.ipaddr = ip
        self.macaddr = mac
        self.mac_manuf = None

    def __str__(self):
        return str(self.hostname or self.ipaddr or macmanuf(self.macaddr))

    def __lt__(self, x):
        if isinstance(x, Host):
            return str(self.ipaddr) < str(x.ipaddr)
        return True

    @property
    def hostname(self):
        return Host.dns.get(str(self.ipaddr))

def load_consts(outdict, module, attrprefix):
    for k in dir(module):
        if k.startswith(attrprefix):
            v = getattr(module, k)
            outdict[v] = k[len(attrprefix):]

def getTuple(pkt):
    if getattrdeep(pkt, 'ip.tcp', None):
        tup = ('tcp', Host.get_host(pkt, 'src'), pkt.ip.tcp.sport, Host.get_host(pkt, 'dst'), pkt.ip.tcp.dport)
    elif getattrdeep(pkt, 'ip.udp', None):
        tup = ('udp', Host.get_host(pkt, 'src'), pkt.ip.udp.sport, Host.get_host(pkt, 'dst'), pkt.ip.udp.dport)
    else:
        return None
    a,b,c,d,e = tup
    if b > d:
        return a,d,e,b,c  # swap src/sport and dst/dport
    else:
        return tup

def getService(tup):
    if not tup: return
    transport, _, sport, _, dport = tup
    if (transport, dport) in services:
        return services.get((transport, dport))
    if (transport, sport) in services:
        return services.get((transport, sport))


def get_transport(pkt):
    ret = 'ether'
    if getattr(pkt, 'arp', None):
        return 'arp'

    if getattr(pkt, 'ip', None):
        ret = 'ip'
        if getattr(pkt.ip, 'tcp', None):
            ret = 'tcp'
        elif getattr(pkt.ip, 'udp', None):
            ret = 'udp'
        elif getattr(pkt.ip, 'icmp', None):
            ret = 'icmp'

    if getattr(pkt, 'ip6', None):
        ret = 'ipv6'
        if getattr(pkt.ip6, 'tcp', None):
            ret = 'tcp'
        elif getattr(pkt.ip6, 'udp', None):
            ret = 'udp'
        elif getattr(pkt.ip6, 'icmp6', None):
            ret = 'icmpv6'

    return ret


def get_port(pkt, field='sport'):
    return getattrdeep(pkt, 'ip.tcp.'+field, None) or getattrdeep(pkt, 'ip.udp.'+field, None)


class EtherSheet(Sheet):
    'Layer 2 (ethernet) packets'
    rowtype = 'packets'
    columns = [
        ColumnAttr('timestamp', type=date, fmtstr="%H:%M:%S.%f"),
        Column('ether_manuf', type=str, getter=lambda col,row: mac_manuf(macaddr(row.src))),
        Column('ether_src', type=str, getter=lambda col,row: macaddr(row.src), width=6),
        Column('ether_dst', type=str, getter=lambda col,row: macaddr(row.dst), width=6),
        ColumnAttr('ether_data', 'data', type=vlen, width=0),
    ]


@VisiData.api
class IPSheet(Sheet):
    rowtype = 'packets'
    columns = [
        ColumnAttr('timestamp', type=date, fmtstr="%H:%M:%S.%f"),
        ColumnAttr('ip', type=str, width=0),
        Column('ip_src', type=str, width=14, getter=lambda col,row: ipaddress.ip_address(row.ip.src)),
        Column('ip_dst', type=str, width=14, getter=lambda col,row: ipaddress.ip_address(row.ip.dst)),
        ColumnAttr('ip_hdrlen', 'ip.hl', type=int, width=0, helpstr="IPv4 Header Length"),
        ColumnAttr('ip_proto', 'ip.p', type=lambda v: protocols['ip'].get(v), width=8, helpstr="IPv4 Protocol"),
        ColumnAttr('ip_id', 'ip.id', type=int, width=10, helpstr="IPv4 Identification"),
        ColumnAttr('ip_rf', 'ip.rf', type=int, width=10, helpstr="IPv4 Reserved Flag (Evil Bit)"),
        ColumnAttr('ip_df', 'ip.df', type=int, width=10, helpstr="IPv4 Don't Fragment flag"),
        ColumnAttr('ip_mf', 'ip.mf', type=int, width=10, helpstr="IPv4 More Fragments flag"),
        ColumnAttr('ip_tos', 'ip.tos', width=10, type=FlagGetter('ip_tos'), helpstr="IPv4 Type of Service"),
        ColumnAttr('ip_ttl', 'ip.ttl', type=int, width=10, helpstr="IPv4 Time To Live"),
        ColumnAttr('ip_ver', 'ip.v', type=int, width=10, helpstr="IPv4 Version"),
    ]

    def iterload(self):
        for pkt in Progress(self.source.rows):
            if getattr(pkt, 'ip', None):
                yield pkt


@VisiData.api
class TCPSheet(IPSheet):
    columns = IPSheet.columns + [
        ColumnAttr('tcp_srcport', 'ip.tcp.sport', type=int, width=8, helpstr="TCP Source Port"),
        ColumnAttr('tcp_dstport', 'ip.tcp.dport', type=int, width=8, helpstr="TCP Dest Port"),
        ColumnAttr('tcp_opts', 'ip.tcp.opts', width=0),
        ColumnAttr('tcp_flags', 'ip.tcp.flags', type=FlagGetter('tcp'), helpstr="TCP Flags"),
    ]

    def iterload(self):
        for pkt in Progress(self.source.rows):
            if getattrdeep(pkt, 'ip.tcp', None):
                yield pkt

class UDPSheet(IPSheet):
    columns = IPSheet.columns + [
        ColumnAttr('udp_srcport', 'ip.udp.sport', type=int, width=8, helpstr="UDP Source Port"),
        ColumnAttr('udp_dstport', 'ip.udp.dport', type=int, width=8, helpstr="UDP Dest Port"),
        ColumnAttr('ip.udp.data', type=vlen, width=0),
        ColumnAttr('ip.udp.ulen', type=int, width=0),
    ]

    def iterload(self):
        for pkt in Progress(self.source.rows):
            if getattrdeep(pkt, 'ip.udp', None):
                yield pkt


class PcapSheet(Sheet):
    rowtype = 'packets'
    columns = [
        ColumnAttr('timestamp', type=date, fmtstr="%H:%M:%S.%f"),
        Column('transport', type=get_transport, width=5),
        Column('srcmanuf', type=str, getter=lambda col,row: manuf(macaddr(row.src))),
        Column('srchost', type=str, getter=lambda col,row: row.srchost),
        Column('srcport', type=int, getter=lambda col,row: get_port(row, 'sport')),
        Column('dstmanuf', type=str, getter=lambda col,row: manuf(macaddr(row.dst))),
        Column('dsthost', type=str, getter=lambda col,row: row.dsthost),
        Column('dstport', type=int, getter=lambda col,row: get_port(row, 'dport')),
        ColumnAttr('ether_proto', 'type', type=lambda v: protocols['ethernet'].get(v), width=0),
        ColumnAttr('tcp_flags', 'ip.tcp.flags', type=FlagGetter('tcp'), helpstr="TCP Flags"),
        Column('service', type=str, getter=lambda col,row: getService(getTuple(row))),
        ColumnAttr('data', type=vlen),
        ColumnAttr('ip_len', 'ip.len', type=int),
        ColumnAttr('tcp', 'ip.tcp', width=4, type=vlen),
        ColumnAttr('udp', 'ip.udp', width=4, type=vlen),
        ColumnAttr('icmp', 'ip.icmp', width=4, type=vlen),
        ColumnAttr('dns', type=str, width=4),
    ]

    def iterload(self):
        init_pcap()

        self.pcap = read_pcap(self.source)
        self.rows = []
        with Progress(total=filesize(self.source)) as prog:
            for ts, buf in self.pcap:
                eth = dpkt.ethernet.Ethernet(buf)
                yield eth
                prog.addProgress(len(buf))

                eth.timestamp = ts
                if not getattr(eth, 'ip', None):
                    eth.ip = getattr(eth, 'ip6', None)
                eth.dns = try_apply(lambda eth: dnslib.DNSRecord.parse(eth.ip.udp.data), eth)
                if eth.dns:
                    for rr in eth.dns.rr:
                        Host.dns[str(rr.rdata)] = str(rr.rname)

                eth.srchost = Host.get_host(eth, 'src')
                eth.dsthost = Host.get_host(eth, 'dst')




flowtype = collections.namedtuple('flow', 'packets transport src sport dst dport'.split())

@VisiData.api
class PcapFlowsSheet(Sheet):
    rowtype = 'netflows'  # rowdef: flowtype
    _rowtype = flowtype

    columns = [
        ColumnAttr('transport', type=str),
        Column('src', type=str, getter=lambda col,row: row.src),
        ColumnAttr('sport', type=int),
        Column('dst', type=str, getter=lambda col,row: row.dst),
        ColumnAttr('dport', type=int),
        Column('service', type=str, width=8, getter=lambda col,row: getService(getTuple(row.packets[0]))),
        ColumnAttr('packets', type=vlen),
        Column('connect_latency_ms', type=float, getter=lambda col,row: col.sheet.latency[getTuple(row.packets[0])]),
    ]

    def iterload(self):
        self.flows = {}
        self.latency = {}  # [flowtuple] -> float ms of latency
        self.syntimes = {}  # [flowtuple] -> timestamp of SYN
        flags = FlagGetter('tcp')
        for pkt in Progress(self.source.rows):
            tup = getTuple(pkt)
            if tup:
                flowpkts = self.flows.get(tup)
                if flowpkts is None:
                    flowpkts = self.flows[tup] = []
                    yield flowtype(flowpkts, *tup)
                flowpkts.append(pkt)

                if not getattr(pkt.ip, 'tcp', None):
                    continue

                tcpfl = flags(pkt.ip.tcp.flags)
                if 'SYN' in tcpfl:
                    if 'ACK' in tcpfl:
                        if tup in self.syntimes:
                            self.latency[tup] = (pkt.timestamp - self.syntimes[tup])*1000
                    else:
                        self.syntimes[tup] = pkt.timestamp

    def openRow(self, row):
        return PcapSheet("%s_packets"%flowname(row), rows=row.packets)


def flowname(flow):
    return '%s_%s:%s-%s:%s' % (flow.transport, flow.src, flow.sport, flow.dst, flow.dport)

def try_apply(func, *args, **kwargs):
    try:
        return func(*args, **kwargs)
    except Exception as e:
        pass


PcapSheet.addCommand('W', 'flows', 'vd.push(PcapFlowsSheet(sheet.name+"_flows", source=sheet))')
PcapSheet.addCommand('2', 'l2-packet', 'vd.push(IPSheet("L2packets", source=sheet))')
PcapSheet.addCommand('3', 'l3-packet', 'vd.push(TCPSheet("L3packets", source=sheet))')

vd.addMenuItem('View', 'Packet capture', 'flows', 'flows')
vd.addMenuItem('View', 'Packet capture', 'IP (L2)', 'l2-packet')
vd.addMenuItem('View', 'Packet capture', 'TCP (L3)', 'l3-packet')



import io

from visidata import AttrDict, vd, Path, asyncthread, Sheet


class GitSheet(Sheet):
    @property
    def gitargstr(self):
        return ' '.join(self.gitargs)

    def git(self, subcmd, *args, **kwargs):
        'For non-modifying commands; not logged except in debug mode'
        sh = vd.importExternal('sh')
        args = list(subcmd.split()) + list(args)
        vd.debug('git ' + ' '.join(str(x) for x in args))
        return sh.git(*args,
                      _cwd=self.gitRootPath,
                      **kwargs)

    def loggit(self, subcmd, *args, **kwargs):
        'Run git command with *args*, and post a status message.'
        import sh
        args = list(subcmd.split()) + list(args)
        vd.warning('git ' + ' '.join(str(x) for x in args))
        return sh.git(*args,
                      _cwd=self.gitRootPath,
                      **kwargs)

    def git_all(self, *args, **kwargs):
        'Return entire output of git command.'
        sh = vd.importExternal('sh')
        try:
            vd.debug('git ' + ' '.join(str(x) for x in args))
            out = self.git('--no-pager',
                      *args,
                      _decode_errors='replace',
                      _bg_exc=False,
                      **kwargs)
        except sh.ErrorReturnCode as e:
            vd.warning('git '+' '.join(str(x) for x in args), 'error=%s' % e.exit_code)
            out = e.stdout

        return out

    def git_lines(self, subcmd, *args, **kwargs):
        'Generator of stdout lines from given git command'
        sh = vd.importExternal('sh')
        err = io.StringIO()
        args = list(subcmd.split()) + list(args)
        try:
            vd.debug('git ' + ' '.join(str(x) for x in args))
            for line in self.git('--no-pager',
                            *args,
                            _decode_errors='replace',
                            _iter=True,
                            _bg_exc=False,
                            _err=err,
                            **kwargs):
                yield line[:-1]  # remove EOL

        except sh.ErrorReturnCode as e:
            vd.warning('git '+' '.join(str(x) for x in args), 'error=%s' % e.exit_code)

        errlines = err.getvalue().splitlines()
        if errlines:
            vd.warning('git stderr: ' + '\n'.join(errlines))


    def git_iter(self, subcmd, *args, sep='\0', **kwargs):
        'Generator of chunks of stdout from given git command *subcmd*, delineated by sep character.'
        sh = vd.importExternal('sh')
        import sh
        err = io.StringIO()

        args = list(subcmd.split()) + list(args)
        bufsize = 512
        chunks = []
        try:
            vd.debug('git ' + ' '.join(str(x) for x in args))
            for data in self.git('--no-pager',
                               *args,
                               _decode_errors='replace',
                               _out_bufsize=bufsize,
                               _iter=True,
                               _bg_exc=False,
                               _err=err,
                               **kwargs):
                while True:
                    i = data.find(sep)
                    if i < 0:
                        break
                    chunks.append(data[:i])
                    data = data[i+1:]
                    yield ''.join(chunks)
                    chunks.clear()

                chunks.append(data)
        except sh.ErrorReturnCode as e:
            vd.warning('git '+' '.join(str(x) for x in args), 'error=%s' % e.exit_code)

        if chunks:
            yield ''.join(chunks)

        errlines = err.getvalue().splitlines()
        if errlines:
            vd.warning('git stderr: ' + '\n'.join(errlines))

    @asyncthread
    def modifyGit(self, *args, **kwargs):
        'Run git command that modifies the repo'
        vd.warning('git ' + ' '.join(str(x) for x in args))
        ret = self.git_all(*args, **kwargs)
        vd.status(ret)

        if isinstance(self.source, GitSheet):
            self.source.reload()

        self.reload()

    @property
    def gitRootSheet(self):
        if isinstance(self.source, GitSheet):
            return self.source.gitRootSheet
        return self

    def iterload(self):
        for line in self.git_lines(*self.gitargs):
            yield AttrDict(line=line)


@GitSheet.lazy_property
def gitRootPath(self):
    'Return Path of git root (nearest ancestor directory with a .git/)'
    def _getRepoPath(p):
        'Return path at p or above which has .git subdir'
        if p.joinpath('.git').exists():
            return p
        if str(p) in ['/','']:
            return None
        return _getRepoPath(p.resolve().parent)

    p = _getRepoPath(self.gitRootSheet.source)
    if p:
       return p


@GitSheet.lazy_property
def branch(self):
    return self.git('rev-parse', '--abbrev-ref', 'HEAD').strip()


GitSheet.options.disp_note_none = ''
GitSheet.options.disp_status_fmt = '{sheet.progressStatus}‹{sheet.branchStatus}› {sheet.name}| '

GitSheet.addCommand('gi', 'git-exec', 'cmdstr=input("gi", type="git"); vd.push(GitSheet(cmdstr, gitargs=cmdstr.split()))', 'execute git command')

GitSheet.addCommand('Alt+g', 'menu-git', 'pressMenu("Git")', '')

vd.addMenuItems('''
    Git > Execute command > git-exec
''')



from typing import Optional, Union
import textwrap

from visidata import vd, VisiData, BaseSheet, colors, TextSheet, clipdraw, wraptext, dispwidth, AttrDict, wrmap
from visidata import CommandHelpGetter, OptionHelpGetter


vd.option('disp_sidebar', True, 'whether to display sidebar')
vd.option('disp_sidebar_fmt', '', 'format string for default sidebar')
vd.theme_option('disp_sidebar_width', 0, 'max width for sidebar')
vd.theme_option('disp_sidebar_height', 0, 'max height for sidebar')
vd.theme_option('color_sidebar', 'black on 114 blue', 'base color of sidebar')
vd.theme_option('color_sidebar_title', 'black on yellow', 'color of sidebar title')

@VisiData.api
class AddedHelp:
    '''Context manager to add help text/screen to list of available sidebars.'''
    def __init__(self, text:Union[str,'HelpPane'], title='', help_flag=''):
        self.helpfunc = None
        if text:
            if not vd.wantsHelp(help_flag):
                return
            self.helpfunc = lambda: (text, title)

    def __enter__(self):
        if self.helpfunc:
            vd._help_sidebars.insert(0, self.helpfunc)
            vd.clearCaches()
        return self

    def __exit__(self, *args):
        if self.helpfunc:
            vd._help_sidebars.remove(self.helpfunc)
            vd.clearCaches()


@BaseSheet.lazy_property
def formatter_helpstr(sheet):
    return AttrDict(commands=CommandHelpGetter(type(sheet)),
                    options=OptionHelpGetter())


@BaseSheet.cached_property
def default_sidebar(sheet):
    'Default to format options.disp_sidebar_fmt.  Overridable.'
    fmt = sheet.options.disp_sidebar_fmt
    if not fmt: return ''
    return sheet.formatString(fmt, help=sheet.formatter_helpstr)


@VisiData.api
def cycleSidebar(vd, n:int=1):
    if vd.sheet.help_sidebars:
        vd.disp_help += n
        if vd.disp_help >= len(vd.sheet.help_sidebars):
            vd.disp_help = -1

        vd.options.disp_sidebar = (vd.disp_help >= 0)
    else:
        vd.disp_help = -1
    vd.clearCaches()


@BaseSheet.cached_property
def help_sidebars(sheet) -> 'list[Callable[[], tuple[str,str]]]':
    r = []
    if sheet.default_sidebar:
        r.append(lambda: (sheet.default_sidebar, ''))
    if sheet.guide:
        r.append(lambda: (sheet.formatString(sheet.guide, help=sheet.formatter_helpstr), ''))
    return r + vd._help_sidebars


@VisiData.cached_property
def sidebarStatus(vd) -> str:
    if vd.sheet.help_sidebars:
        if vd.wantsHelp('sidebar') and vd.disp_help >= 0:
            n = vd.disp_help+1
            return f'[:onclick sidebar-toggle][:sidebar][{n}/{len(vd.sheet.help_sidebars)}][/]'
        else:
            return f'[:onclick sidebar-toggle][:sidebar][{len(vd.sheet.help_sidebars)}][/]'

    return ''

@VisiData.property
def recentStatusMessages(vd) -> str:
    r = ''
    for (pri, msgparts), n in vd.statuses.items():
        msg = '; '.join(wrmap(str, msgparts))
        msg = f'[{n}x] {msg}' if n > 1 else msg

        if pri == 3: msgattr = '[:error]'
        elif pri == 2: msgattr = '[:warning]'
        elif pri == 1: msgattr = '[:warning]'
        else: msgattr = ''

        if msgattr:
            r += '\n' + f'{msgattr}{msg}[/]'
        else:
            r += '\n' + msg

    if r:
        return '# statuses' + r

    return ''


@VisiData.api
def drawSidebar(vd, scr, sheet):
    sidebar = vd.recentStatusMessages
    title = ''
    bottommsg = ''
    overflowmsg = '[:reverse] Ctrl+P to view all status messages [/]'
    try:
        if not sidebar and vd.options.disp_sidebar and vd.wantsHelp('guides') and sheet.help_sidebars:
            sidebar, title = sheet.help_sidebars[vd.disp_help%len(sheet.help_sidebars)]()

#            bottommsg = sheet.formatString('[:onclick sidebar-toggle][:reverse] {help.commands.sidebar_toggle} [:]', help=sheet.formatter_helpstr)
#            overflowmsg = '[:reverse] see full sidebar with [:code]gb[/] [:]'

            overflowmsg = '[:onclick open-sidebar]…↓…[/]'
    except Exception as e:
        vd.exceptionCaught(e)
        sidebar = f'# error\n{e}'

    sheet.current_sidebar = sidebar

    return sheet.drawSidebarText(scr, text=sheet.current_sidebar, title=title, overflowmsg=overflowmsg, bottommsg=bottommsg)

@BaseSheet.api
def drawSidebarText(sheet, scr, text:Union[None,str,'HelpPane'], title:str='', overflowmsg:str='', bottommsg:str=''):
    scrh, scrw = scr.getmaxyx()
    maxw = sheet.options.disp_sidebar_width or scrw//2
    maxh = sheet.options.disp_sidebar_height or max(scrh-2, 1)

    cattr = colors.get_color('color_sidebar')

    text = text or ''

    if hasattr(text, 'draw'):  # like a HelpPane
        maxlinew = text.width
        winh = min(maxh, text.height+2)+1
    else:
        text = textwrap.dedent(text.strip('\n'))

        if not text:
            return

        lines = text.splitlines()
        if not title and lines and lines[0].strip().startswith('# '):
            title = lines[0].strip()[2:]
            text = '\n'.join(lines[1:])

        if not text:
            return

        lines = list(wraptext(text, width=maxw-4))
        maxlinew = 0
        if lines:
            maxlinew = max(maxlinew, max(dispwidth(textonly, maxwidth=maxw) for line, textonly in lines))
        winh = min(maxh, len(lines)+2)
    winh = max(winh, 1)

    titlew = dispwidth(title)

    maxlinew = max(maxlinew, dispwidth(overflowmsg)+4)
    maxlinew = max(maxlinew, dispwidth(bottommsg)+4)
    maxlinew = max(maxlinew, titlew)
    winw = min(maxw, maxlinew+4)
    x, y, w, h = scrw-winw-1, scrh-winh-1, winw, winh
    y = max(y, 0)

    sidebarscr = vd.subwindow(scr, x, y, w, h)

    sidebarscr.erase()
    sidebarscr.bkgd(' ', cattr.attr)
    sidebarscr.border()
    vd.onMouse(sidebarscr, 0, 0, w, h, BUTTON1_RELEASED='no-op', BUTTON1_PRESSED='no-op')

    if hasattr(text, 'draw'):  # like a HelpPane
        text.draw(sidebarscr, attr=cattr)
    else:
        i = 0
        for line, _ in lines:
            if i >= h-2:
                bottommsg = overflowmsg
                break

            x += clipdraw(sidebarscr, i+1, 2, line, cattr, w=w-3)
            i += 1

    x = max(0, w-titlew-6)
    clipdraw(sidebarscr, 0, x, f"|[:sidebar_title] {title} [:]|", cattr, w=titlew+4)
    if bottommsg:
        clipdraw(sidebarscr, h-1, winw-dispwidth(bottommsg)-4, '|'+bottommsg+'|', cattr)

    sidebarscr.noutrefresh()


@VisiData.api
class SidebarSheet(TextSheet):
    guide = '''
        # Sidebar Guide
        The sidebar provides additional information about the current sheet, defaulting to a basic guide to the current sheet type.
        It can be configured to show many useful attributes via `options.disp_sidebar_fmt`.

        - `gb` to open the sidebar in a new sheet
        - `b` to toggle the sidebar on/off for the current sheet
    '''

BaseSheet.addCommand('b', 'sidebar-toggle', 'vd.options.disp_sidebar = not vd.options.disp_sidebar', 'toggle sidebar')
BaseSheet.addCommand('gb', 'open-sidebar', 'sheet.current_sidebar = "" if not hasattr(sheet, "current_sidebar") else sheet.current_sidebar; vd.push(SidebarSheet(name, options.disp_sidebar_fmt, source=sheet.current_sidebar.splitlines()))', 'open sidebar in new sheet')
BaseSheet.addCommand('^G', 'sidebar-cycle', 'vd.cycleSidebar()', 'cycle through available sidebar panels')


vd.addMenuItems('''
    View > Sidebar > toggle > sidebar-toggle
    View > Sidebar > cycle > sidebar-cycle
    View > Sidebar > open in new sheet > open-sidebar
''')



import os

from visidata import VisiData, vd, Column, TableSheet, vlen

@VisiData.api
def open_eml(vd, p):
    return EmailSheet(p.base_stem, source=p)

open_mhtml = open_eml

class EmailSheet(TableSheet):
    rowtype = 'parts'  # rowdef: sub-Messages
    columns = [
        Column('filename', getter=lambda c,r: r.get_filename()),
        Column('content_type', getter=lambda c,r: r.get_content_type()),
        Column('payload', type=vlen, getter=lambda c,r: r.get_payload(decode=False)),
    ]
    def iterload(self):
        import email.parser
        parser = email.parser.Parser()
        with self.source.open(encoding='utf-8') as fp:
            yield from parser.parse(fp).walk()

@EmailSheet.api
def extract_part(sheet, givenpath, part):
    payload = part.get_payload(decode=True)
    if payload is None:
        vd.warning('empty payload')
    else:
        with givenpath.open_bytes(mode='w') as fp:
            fp.write(payload)

@EmailSheet.api
def extract_parts(sheet, givenpath, *parts):
    'Save all *parts* to Path *givenpath*.'
    vd.confirmOverwrite(givenpath, f'{givenpath} already exists, extract anyway?')

    vd.status('saving %s parts to %s' % (len(parts), givenpath.given))

    # forcibly specify save individual files into directory by ending path with /
    if givenpath.is_dir() or givenpath.given.endswith('/') or len(parts) > 1:
        # save as individual files in the givenpath directory
        try:
            os.makedirs(givenpath, exist_ok=True)
        except FileExistsError:
            vd.debug(f'{givenpath} already exists')

        for i, part in enumerate(parts):
            fn = part.get_filename() or f'part{i}'
            vd.execAsync(sheet.extract_part, givenpath / fn, part)
    elif len(parts) == 1:
        vd.execAsync(sheet.extract_part, givenpath, parts[0])
    else:
        vd.fail('cannot save multiple parts to non-dir')

EmailSheet.addCommand('x', 'extract-part', 'extract_part(inputPath("save part as: ", value=cursorRow.get_filename()), cursorRow)')
EmailSheet.addCommand('gx', 'extract-part-selected', 'extract_parts(inputPath("save %d parts in: " % nSelectedRows), *selectedRows)')



import pytest
import visidata

def pytest_generate_tests(metafunc):
    """Split feature tests into separate test cases

    Look up test methods in imported modules. Turn each one into a single test_feature()
    case, with "module::method" as the test id.
    """
    tests = [
        (mod, getattr(mod, k))
        for mod in visidata.vd.importedModules
        for k in dir(mod)
        if k.startswith("test_")
    ]
    argvalues = [[testfunc] for _, testfunc in tests]
    testids = [
        f"{mod.__name__}::{testfunc.__name__}"
        for mod, testfunc in tests
    ]
    metafunc.parametrize(argnames=["testfunc"], argvalues=argvalues, ids=testids)


@pytest.mark.usefixtures("curses_setup")
def test_feature(mock_screen, testfunc):
    visidata.vd.resetVisiData()
    visidata.vd.scr = mock_screen
    testfunc(visidata.vd)



from copy import copy
from visidata import vd, Sheet, Progress, asyncthread, options, UNLOADED

@Sheet.api
@asyncthread
def sortSelected(self, ordering):
    """sort the rows that are selected, in place so that each originally-selected row takes the place of another.
    *ordering* is a list of tuples: (col_name_str, reverse_sort_bool) (string and bool)"""
    if self.rows is UNLOADED or not self.rows:
        return
    if self.nSelectedRows == 0:
        vd.fail('no rows selected')
        return

    if options.undo:
        vd.status('undo added')
        vd.addUndo(setattr, self, 'rows', copy(self.rows))

# temporary:  save data to test integrity after the sort
    rows_initial = { self.rowid(r) for r in self.rows }
    selected_initial = set(self._selectedRows.keys())

    with Progress(gerund='sorting', total=self.nSelectedRows) as prog:
        selected = list(self._selectedRows.values())
        selected_rowids = set(self._selectedRows.keys())

        selected_indices = []
        for i, r in enumerate(self.rows):
            rowid = self.rowid(r)
            if rowid in selected_rowids:
                selected_indices.append(i)
                selected_rowids.remove(rowid)
            if len(selected_rowids) == 0:
                break
        try:
            def _sortkey(r):
                prog.addProgress(1)
                return self.sortkey(r, ordering=ordering)
            sorted_selected = sorted(selected, key=_sortkey)
        except TypeError as e:
            vd.warning('sort incomplete due to TypeError; change column type')
            vd.exceptionCaught(e, status=False)
            return
        for selected_idx, r in zip(selected_indices, sorted_selected, strict=True):
            self.rows[selected_idx] = r

# temporary:  make sure we haven't lost any rows in the sheet or the selection
    rows_final = { self.rowid(r) for r in self.rows }
    selected_final = set(self._selectedRows.keys())
    assert(rows_initial == rows_final)
    assert(selected_initial == selected_final)

Sheet.addCommand(None, 'sort-selected-asc', 'sortSelected([(cursorCol, False)])', 'sort selected rows by the current column, in ascending order')
Sheet.addCommand(None, 'sort-selected-desc', 'sortSelected([(cursorCol, True)])', 'sort selected rows by the current column, in descending order')



import ast

from visidata import vd, Column, ExprColumn, Sheet

vd.option('rename_cascade', False, 'cascade column renames into expressions')  #2088


class Renamer(ast.NodeTransformer):
    def __init__(self, find, replace):
        self.find = find
        self.replace = replace

    def visit_Name(self, node):
        if node.id == self.find:
            node.id = self.replace

        return node


@Column.before
def setName(col, newname):
    if col.sheet and col.sheet.options.rename_cascade:
        for c in col.sheet.columns:
            if isinstance(c, ExprColumn):
                parsed_expr = ast.parse(c.expr)
                canon_expr = ast.unparse(parsed_expr)
                new_expr = ast.unparse(Renamer(col.name, newname).visit(parsed_expr))
                if new_expr != canon_expr:
                    vd.addUndo(setattr, c, 'expr', c.expr)
                    c.expr = new_expr



import re
from visidata import vd, VisiData, Sheet


vd.option('clean_names', False, 'clean column/sheet names to be valid Python identifiers', replay=True)


@VisiData.global_api
def cleanName(vd, s):
    #[Nas Banov] https://stackoverflow.com/a/3305731
    #    return re.sub(r'\W|^(?=\d)', '_', str(s)).strip('_')
    s = re.sub(r'[^\w\d_]', '_', s)  # replace non-alphanum chars with _
    s = re.sub(r'_+', '_', s)  # replace runs of _ with a single _
    s = s.strip('_')
    return s


@Sheet.api
def maybeClean(sheet, s):
    if sheet.options.clean_names:
        s = vd.cleanName(s)
    return s


Sheet.addCommand('', 'clean-names', '''
options.clean_names = True;
for c in visibleCols:
    c.name = cleanName(c.name)
''', 'set options.clean_names on sheet and clean visible column names')



from visidata import vd, VisiData, ItemColumn, AttrDict, RowColorizer, Path

from .gitsheet import GitSheet

@VisiData.api
def git_remote(vd, p, args):
    if not args or 'show' in args:
        return GitRemotes('remotes', source=p)


class GitRemotes(GitSheet):
    guide = '''
        # git remote
        Manage the set of repositories ("remotes") whose branches you track.

        - `a` to add a remote
        - `d` to mark a remote for deletion
        - `e` to edit the _remote_ or _url_
        - `z Ctrl+S` to commit the changes.
    '''
    rowtypes = 'remotes'  # rowdef: dict(remote=, url=, type=)
    columns=[
        ItemColumn('remote', setter=lambda c,r,v: c.sheet.set_remote(c,r,v)),
        ItemColumn('type'),
        ItemColumn('url', width=40, setter=lambda c,r,v: c.sheet.set_url(c,r,v)),
    ]
    nKeys = 1
    defer = True

    def set_remote(self, col, row, val):
        self.loggit('remote', 'rename', self.column('remote').getSourceValue(row), val)

    def set_url(self, col, row, val):
        self.loggit('remote', 'set-url', row.remote, val)

    def iterload(self):
        for line in self.git_lines('remote', '-v', 'show'):
            name, url, paren_type = line.split()
            yield AttrDict(remote=name, url=url, type=paren_type[1:-1])

    def commitDeleteRow(self, row):
        self.loggit('remote', 'remove', row.remote)

    def commitAddRow(self, row):
        row.remote = self.column('remote').getValue(row)
        row.url = self.column('url').getValue(row)
        self.loggit('remote', 'add', row.remote, row.url)

    def newRow(self):
        return AttrDict()


GitSheet.addCommand('', 'git-open-remotes', 'vd.push(git_remote(Path("."), ""))', 'open git remotes sheet')

vd.addMenuItems('''
    Git > Open > remotes > git-open-remotes
''')



from functools import reduce
from copy import deepcopy

import json

from visidata import VisiData, vd, Column, asyncthread, Progress, PythonSheet, InvertedCanvas, date, wrapply, TypedExceptionWrapper, TypedWrapper



@VisiData.api
def open_geojson(vd, p):
    return GeoJSONSheet(p.base_stem, source=p)

class GeoJSONColumn(Column):
    def calcValue(self, row):
        return row.get('properties', {}).get(self.expr)

    def putValue(self, row, val):
        properties = row.setdefault('properties', {})
        properties[self.expr] = val


class GeoJSONSheet(PythonSheet):
    rowtype = 'shapes'
    columns = [Column('json_row', width=0)]

    def iterload(self):
        self.colnames = {}

        with self.source.open(encoding='utf-8') as fp:
            ret = json.load(fp)

            if ret['type'] == 'FeatureCollection':
                features = ret['features']
            elif ret['type'] == 'Feature':
                features = [ret]
            elif ret['type'] == 'GeometryCollection':
                features = list(map(lambda g: { 'type': 'Feature', 'geometry': g }, ret['geometries']))
            else: # Some form of geometry
                features = [{ 'type': 'Feature', 'geometry': ret }]

            for feature in Progress(features):
                for prop in feature.get('properties', {}).keys():
                    prop = self.maybeClean(prop)
                    if prop not in self.colnames:
                        c = GeoJSONColumn(name=prop, expr=prop)
                        self.colnames[prop] = c
                        self.addColumn(c)
                yield feature

class GeoJSONMap(InvertedCanvas):
    aspectRatio = 1.0
    filetype = 'geojson'

    @asyncthread
    def reload(self):
        self.reset()

        nplotted = nerrors = 0

        for row in Progress(self.sourceRows):
            k = self.source.rowkey(row)
            colour = self.plotColor(k)

            try:
                bbox = self.parse_geometry(row, colour)
                nplotted += 1
            except Exception as e:
                vd.exceptionCaught(e)
                nerrors += 1
                continue

            x1, y1, x2, y2 = bbox
            textx, texty = (x1+x2)/2, (y1+y2)/2
            disptext = self.textCol.getDisplayValue(row)
            self.label(textx, texty, disptext, colour, row)

        vd.status('loaded %d %s (%d errors)' % (nplotted, self.rowtype, nerrors))
        self.refresh()

    def parse_geometry(self, row, colour, bbox=None):
        if bbox is None: bbox = [180, 90, -180, -90]

        typ = row['geometry']['type']
        if typ == 'GeometryCollection':
            for g in row['geometries']:
                bbox = self.parse_geometry(row, colour, bbox)
            return bbox

        coords = row['geometry']['coordinates']
        if typ in ('Point', 'LineString', 'Polygon'):
            coords = [coords]

        if typ in ('Point', 'MultiPoint'):
            for x, y in coords:
                self.point(x, y, colour, row)
            bbox = reduce_coords(coords, bbox)
        elif typ in ('LineString', 'MultiLineString'):
            for line in coords:
                self.polyline(line, colour, row)
                bbox = reduce_coords(line, bbox)
        elif typ in ('Polygon', 'MultiPolygon'):
            for polygon in coords:
                if not isinstance(polygon[0][0], list):
                    continue
                self.polygon(polygon[0], colour, row)
                bbox = reduce_coords(polygon[0], bbox)
                for hole in polygon[1:]:
                    self.polygon(hole, 0, row)
        else:
            vd.warning('notimpl shapeType %s' % typ)

        return bbox

def reduce_coords(coords, initial):
    return reduce(
        lambda a,n: [min(a[0],n[0]), min(a[1],n[1]), max(a[2],n[0]), max(a[3],n[1])],
        coords, initial)

def _rowdict(cols, row):
    ret = {}
    for col in cols:
        o = wrapply(col.getTypedValue, row)
        if isinstance(o, TypedExceptionWrapper):
            o = col.sheet.options.safe_error or str(o.exception)
        elif isinstance(o, TypedWrapper):
            o = o.val
        elif isinstance(o, date):
            o = col.getDisplayValue(row)
        if o is not None:
            ret[col.name] = o
    return ret

@VisiData.api
def save_geojson(vd, p, vs):
    features = []
    for row in Progress(vs.rows, 'saving'):
        copyrow = deepcopy(row)
        copyrow['properties'] = _rowdict(vs.visibleCols, row)
        features.append(copyrow)

    featcoll = {
        'type': 'FeatureCollection',
        'features': features,
    }

    try:
        indent = int(vs.options.json_indent)
    except Exception:
        indent = vs.options.json_indent

    with p.open(mode='w', encoding='utf-8') as fp:
        encoder = json.JSONEncoder(indent=indent, sort_keys=vs.options.json_sort_keys)
        for chunk in encoder.iterencode(featcoll):
            fp.write(chunk)

GeoJSONSheet.addCommand('.', 'plot-row', 'vd.push(GeoJSONMap(name+"_map", sourceRows=[cursorRow], textCol=cursorCol, source=sheet))', 'plot geospatial vector in current row')
GeoJSONSheet.addCommand('g.', 'plot-rows', 'vd.push(GeoJSONMap(name+"_map", sourceRows=rows, textCol=cursorCol, source=sheet))', 'plot all geospatial vectors in current sheet')
GeoJSONMap.addCommand('^S', 'save-sheet', 'vd.saveSheets(inputPath("save to: ", value=getDefaultSaveName(sheet)), sheet)', 'save current sheet to filename in format determined by extension (default .geojson)')

vd.addGlobals({
    'GeoJSONMap': GeoJSONMap,
})



"""Allow VisiData to work directly with Amazon S3 paths.

Functionality is more limited than local paths, but supports:

* Navigating among directories (S3 prefixes)
* Opening supported filetypes, including compressed files
* Versioned buckets
"""

import textwrap
from visidata import (
    ENTER,
    Column,
    ItemColumn,
    Path,
    Sheet,
    VisiData,
    asyncthread,
    date,
    vd,
)

vd.option(
    "s3_endpoint",
    "",
    "alternate S3 endpoint, used for local testing or alternative S3-compatible services",
    replay=True,
)
vd.option("s3_glob", True, "enable glob-matching for S3 paths", replay=True)
vd.option(
    "s3_version_aware",
    False,
    "show all object versions in a versioned bucket",
    replay=True,
)


class S3Path(Path):
    """A Path-like object representing an S3 file (object) or directory (prefix)."""

    _fs = None

    def __init__(self, path, version_aware=None, version_id=None):
        super().__init__(path)
        self.given = path
        self.version_aware = version_aware or vd.options.s3_version_aware
        self.version_id = self.version_aware and version_id or None

    @property
    def fs(self):
        if self._fs is None:
            s3fs_core = vd.importExternal("s3fs.core", "s3fs")
            self._fs = s3fs_core.S3FileSystem(
                client_kwargs={"endpoint_url": vd.options.s3_endpoint or None},
                version_aware=self.version_aware,
            )

        return self._fs

    @fs.setter
    def fs(self, val):
        self._fs = val

    def open(self, mode='r', **kwargs):
        """Open the current S3 path, decompressing along the way if needed."""

        fp = self.fs.open(self.given, mode="rb" if self.compression else mode, version_id=self.version_id)

        # Workaround for https://github.com/ajkerrigan/visidata-plugins/issues/12
        if hasattr(fp, "cache") and fp.cache.size != fp.size:
            vd.debug(
                f"updating cache size from {fp.cache.size} to {fp.size} to match object size"
            )
            fp.cache.size = fp.size

        if self.compression == "gz":
            import gzip

            return gzip.open(fp, mode, **kwargs)

        if self.compression == "bz2":
            import bz2

            return bz2.open(fp, mode, **kwargs)

        if self.compression == "xz":
            import lzma

            return lzma.open(fp, mode, **kwargs)

        return fp


class S3DirSheet(Sheet):
    """Display a listing of files and directories (objects and prefixes) in an S3 path.

    Allow single or multiple entries to be opened in separate sheets.
    """

    columns = [
        Column("name", getter=lambda col, row: col.sheet.object_display_name(row)),
        ItemColumn("type"),
        ItemColumn("size", type=int),
        ItemColumn("modtime", "LastModified", type=date),
        ItemColumn("latest", "IsLatest", type=bool),
        ItemColumn("version_id", "VersionId", type=str, width=0),
    ]

    def __init__(self, name, source, version_aware=None):
        import re

        super().__init__(name=name, source=source)
        self.rowtype = "files"
        self.nKeys = 1
        self.use_glob_matching = vd.options.s3_glob and re.search(
            r"[*?\[\]]", self.source.given
        )
        self.version_aware = (
            vd.options.s3_version_aware if version_aware is None else version_aware
        )
        self.fs = source.fs

    def object_display_name(self, row):
        """Provide a friendly display name for an S3 path.

        When listing the contents of a single S3 prefix, the name can chop off
        prefix bits to imitate a directory browser. When glob matching,
        include the full key name for each entry.
        """
        return (
            row.get("name")
            if self.use_glob_matching
            else row.get("name").rpartition("/")[2]
        )

    def iterload(self):
        """Delegate to the underlying filesystem to fetch S3 entries."""
        list_func = self.fs.glob if self.use_glob_matching else self.fs.ls

        if not (
            self.use_glob_matching
            or self.fs.exists(self.source.given)
            or self.fs.isdir(self.source.given)
        ):
            vd.fail(f"unable to open S3 path: {self.source.given}")

        if self.version_aware:
            self.column("latest").hide(False)
        else:
            self.column("latest").hide(True)

        for key in list_func(str(self.source)):
            if self.version_aware and self.fs.isfile(key):
                yield from (
                    {**obj_version, "name": key, "type": "file"}
                    for obj_version in self.fs.object_version_info(key)
                    if key.partition("/")[2] == obj_version["Key"]
                )
            else:
                yield self.fs.stat(key)

    @asyncthread
    def download(self, rows, savepath):
        """Download files and directories to a local path.

        Recurse through through subdirectories.
        """
        remote_files = [row["name"] for row in rows]
        self.fs.download(remote_files, str(savepath), recursive=True)

    def open_rows(self, rows):
        """Open new sheets for the target rows."""
        return (
            vd.openSource(
                S3Path(
                    "s3://{}".format(row["name"]),
                    version_aware=self.version_aware,
                    version_id=row.get("VersionId"),
                )
            )
            for row in rows
        )

    def join_rows(self, rows):
        """Open new sheets for the target rows and concatenate their contents."""
        sheets = list(self.open_rows(rows))
        for sheet in vd.Progress(sheets):
            sheet.reload()

        # Wait for all sheets to fully load before joining them.
        # 'append' is the only join type that makes sense here,
        # since we're joining freshly opened sheets with no key
        # columns.
        vd.sync()
        return sheets[0].openJoin(sheets[1:], jointype="append")

    def refresh_path(self, path=None):
        """Clear the s3fs cache for the given path and reload.

        By default, clear the entire cache.
        """
        self.fs.invalidate_cache(path)
        self.reload()

    def toggle_versioning(self):
        """Enable or disable support for S3 versioning."""
        self.version_aware = not self.version_aware
        self.fs.version_aware = self.version_aware
        vd.status(f's3 versioning {"enabled" if self.version_aware else "disabled"}')
        if self.currentThreads:
            vd.debug("cancelling threads before reloading")
            vd.cancelThread(*self.currentThreads)
        self.reload()


@VisiData.api
def openurl_s3(vd, p, filetype):
    """Open a sheet for an S3 path.

    S3 directories (prefixes) require special handling, but files (objects)
    can use standard VisiData "open" functions.
    """

    # Non-obvious behavior here: For the default case, we don't want to send
    # a custom endpoint to s3fs. However, using None as a default trips up
    # VisiData's type detection for the endpoint option. So we use an empty
    # string as the default instead, and convert back to None here.
    endpoint = vd.options.s3_endpoint or None

    p = S3Path(
        str(p.given),
        version_aware=getattr(p, "version_aware", vd.options.s3_version_aware),
        version_id=getattr(p, "version_id", None),
    )

    p.fs.version_aware = p.version_aware
    if p.fs.client_kwargs.get("endpoint_url", "") != endpoint:
        p.fs.client_kwargs = {"endpoint_url": endpoint}
        p.fs.connect()

    if not p.fs.isfile(str(p.given)):
        return S3DirSheet(p.base_stem, source=p, version_aware=p.version_aware)

    if not filetype:
        filetype = p.ext or "txt"

    openfunc = getattr(vd, f"open_{filetype.lower()}")
    if not openfunc:
        vd.warning(f"no loader found for {filetype} files, falling back to txt")
        filetype = "txt"
        openfunc = vd.open_txt

    assert callable(openfunc), f"no function/method available to open {p.given}"
    vs = openfunc(p)
    vd.status(
        f'opening {p.given} as {filetype} (version id: {p.version_id or "latest"})'
    )
    return vs


S3DirSheet.addCommand(
    ENTER,
    "s3-open-row",
    "vd.push(next(sheet.open_rows([cursorRow])))",
    "open the current S3 entry",
)
S3DirSheet.addCommand(
    "g" + ENTER,
    "s3-open-rows",
    "for vs in sheet.open_rows(selectedRows): vd.push(vs)",
    "open all selected S3 entries",
)
S3DirSheet.addCommand(
    "z^R",
    "s3-refresh-sheet",
    "sheet.refresh_path(str(sheet.source))",
    "clear the s3fs cache for this path, then reload",
)
S3DirSheet.addCommand(
    "gz^R",
    "s3-refresh-sheet-all",
    "sheet.refresh_path()",
    "clear the entire s3fs cache, then reload",
)
S3DirSheet.addCommand(
    "^V",
    "s3-toggle-versioning",
    "sheet.toggle_versioning()",
    "enable/disable support for S3 versioning",
)
S3DirSheet.addCommand(
    "&",
    "s3-join-rows",
    "vd.push(sheet.join_rows(selectedRows))",
    "open and join sheets for selected S3 entries",
)
S3DirSheet.addCommand(
    "gx",
    "s3-download-rows",
    textwrap.dedent(
        """
        savepath = inputPath("download selected rows to: ", value=".")
        sheet.download(selectedRows, savepath)
    """
    ),
    "download selected files and directories",
)

S3DirSheet.addCommand(
    "x",
    "s3-download-row",
    # Note about the use of `_path.name` here. Given a `visidata.Path`
    # object `path`, `path._path` is a `pathlib.Path` object.
    #
    # `visidata.Path` objects do some fun parsing to pick out
    # file types and extensions, handle compression transparently,
    # etc. That parsing leaves the `name` attribute without a file
    # extension, and makes it a little tricky to tack back on.
    #
    # `pathlib.Path` objects have a `name` with the extension intact.
    # That makes `path._path.name` a convenient default output path.
    textwrap.dedent(
        """
        savepath = inputPath("download to: ", value=Path(cursorRow["name"])._path.name)
        sheet.download([cursorRow], savepath)
    """
    ),
    "download the file or directory in the cursor row",
)

vd.addMenuItems(
    """
    File > Toggle versioning > s3-toggle-versioning
    File > Refresh > Current path > s3-refresh-sheet
    File > Refresh > All > s3-refresh-sheet-all
    Row > Download > Current row > s3-download-row
    Row > Download > Selected rows > s3-download-rows
    Data > Join > Selected rows > s3-join-rows
"""
)

vd.addGlobals(S3DirSheet=S3DirSheet)



import string
import textwrap
import time
import curses

from typing import List, Union
from visidata import vd, drawcache, colors, clipdraw, dispwidth
from visidata import BaseSheet, VisiData, AttrDict, ENTER

vd.option('disp_menu', True, 'show menu on top line when not active', sheettype=None)
vd.theme_option('disp_menu_keys', True, 'show keystrokes inline in submenus', sheettype=None)
vd.theme_option('color_menu', 'black on 68 blue', 'color of menu items in general')
vd.theme_option('color_menu_active', '223 yellow on black', 'color of active menu items')
vd.theme_option('color_menu_spec', 'black on 34 green', 'color of sheet-specific menu items')
vd.theme_option('color_menu_help', 'black italic on 68 blue', 'color of helpbox')

vd.theme_option('disp_menu_boxchars', '││──┌┐└┘├┤', 'box characters to use for menus')
vd.theme_option('disp_menu_more', '»', 'command submenu indicator')
vd.theme_option('disp_menu_push', '⎘', 'indicator if command pushes sheet onto sheet stack')
vd.theme_option('disp_menu_input', '…', 'indicator if input required for command')
vd.option('disp_menu_fmt', '| VisiData {vd.version} | {vd.motd}', 'right-side menu format string')

BaseSheet.init('activeMenuItems', list)
vd.menuRunning = False

@VisiData.property
def hintStatus(vd):
    if vd.wantsHelp('hints'):
        if int(time.time()/60) % 2 == 0:
            return 'Alt+H for help menu'
        else:
            return 'Ctrl+G to cycle help sidebar'

    return vd.sheet.getHint()

def menudraw(*args):
    return clipdraw(*args, truncator='')


def Menu(title, *args):
    'Construct menu command or submenu.  *title* is displayed text for this item; *args* is either a single command longname, or recursive Menu elements.'
    if len(args) == 1 and isinstance(args[0], str):
        return AttrDict(title=title, menus=[], longname=args[0])

    return AttrDict(title=title, menus=list(args), longname='')


def walkmenu(item, menupath=[]):
    if item.menus:
        for i in item.menus:
            yield from walkmenu(i, menupath+[i.title])
    else:
        yield item, menupath


def _finditem(menus, item:Union[str,int]):
    if isinstance(item, str):
        for m in menus:
            if item == m.title:
                return m
        return None

    return menus[item]


@BaseSheet.api
def getMenuItem(sheet, menupath:List[str]=None):
    if not menupath:
        menupath = sheet.activeMenuItems

    try:
        currentItem = sheet
        for i in menupath:
            currentItem = _finditem(currentItem.menus, i)
    except IndexError as e:
        vd.exceptionCaught(e)
#        sheet.activeMenuItems = sheet.activeMenuItems[:i-1]

    return currentItem


@VisiData.api
def addMenuItem(vd, *args):
    'Add one command to hierarchical menu at given menupath.  Last argument must be valid command longname. Append menupath elements that do not exist. Example: vd.addMenuItem("File", "Options", "more options", "options-more")'
    m = Menu(*args[-2:])
    for x in reversed(args[:-2]):
        m = Menu(x, m)
    vd.addMenu(m)


@VisiData.api
def addMenuItems(vd, *itemgroups):
    '''Add any number of commands to menu, separated by lines, with individual menupaths separated by '>' character.  Example:
        vd.addMenuItems("""
            Help > About > credits > show-credits
            Help > About > environment > show-env
        """)
    '''
    for itemgroup in itemgroups:
        for itemline in itemgroup.splitlines():
            if not itemline: continue
            menupath = [x.strip() for x in itemline.split('>')]
            vd.addMenuItem(*menupath)


@VisiData.api
def addMenu(vd, *args):
    '''Incorporate submenus and commands into hierarchical menu.  Wrap all in Menu() objects.  Example:

        vd.addMenu(Menu("Help",
                    Menu("About",
                      Menu("credits", "show-credits"),
                      Menu("environment", "show-env"),
                  )))'''
    m = Menu('top', *args)
    for item, menupath in walkmenu(m):
        obj = vd
        for p in menupath:
            c = _finditem(obj.menus, p)
            if not c:
                c = Menu(p)
                obj.menus.append(c)
            obj = c
        assert not obj.menus, 'cannot override submenu with longname'
        obj.longname = item.longname


def _intMenuPath(obj, menupath):
    'Return list of numeric indexes of *menupath* (which may be numeric or string titles) through obj.menus.'
    if not menupath:
        return []

    i = menupath[0]
    if isinstance(i, str):
        try:
            i = [x.title for x in obj.menus].index(i)
        except ValueError:
            vd.warning('no menupath %s' % menupath)
            return []

    return [i] + _intMenuPath(obj.menus[i], menupath[1:])

vd.menus = []

vd.addMenu(
    Menu('File'),
    Menu('Edit'),
    Menu('View'),
    Menu('Column'),
    Menu('Row'),
    Menu('Data'),
    Menu('Plot'),
    Menu('System'),
    Menu('Help')
)


@BaseSheet.api
def menuitemAvailable(sheet, item):
    return any(sheet.menuitemAvailable(i) for i in item.menus)


@VisiData.api
def drawSubmenu(vd, scr, sheet, y, x, menus, level, disp_menu_boxchars=''):
    if not menus:
        return
    ls,rs,ts,bs,tl,tr,bl,br,lsr,rsl = disp_menu_boxchars

    try:
        sheet.activeMenuItems[level] %= len(menus)
    except IndexError:
        pass # sheet.activeMenuItems = sheet.activeMenuItems[:-1]

    for item in menus:
        if item.cmd:
            mainbinding = sheet.revbinds.get(item.cmd.longname, [None])[0]
            item.binding = vd.prettykeys(mainbinding)

    maxbinding = 0
    if vd.options.disp_menu_keys:
        maxbinding = max(dispwidth(item.binding or '') for item in menus)+1

    w = max(dispwidth(item.title) for item in menus)+maxbinding+2

    # draw borders before/under submenus
    if level > 1:
        menudraw(scr, y-1, x, tl+ts*(w+2)+tr, colors.color_menu)  # top

    menudraw(scr, y+len(menus), x, bl+bs*(w+2)+br, colors.color_menu) #  bottom

    i = 0
    for j, item in enumerate(menus):
        attr = colors.color_menu

        if any(foo.obj not in ['BaseSheet', 'TableSheet'] for foo, _ in walkmenu(item)):
            attr = colors.color_menu_spec

        if level < len(sheet.activeMenuItems):
          if j == sheet.activeMenuItems[level]:
            attr = colors.color_menu_active

            if level < len(sheet.activeMenuItems):
                vd.drawSubmenu(scr, sheet, y+i, x+w+4, item.menus, level+1, disp_menu_boxchars=disp_menu_boxchars)

        menudraw(scr, y+i, x, ls, colors.color_menu)

        title = item.title
        pretitle= ' '
        titlenote= ' '
        if item.menus:
            titlenote = vd.options.disp_menu_more

        mainbinding = ''
        # special notes
        if item.cmd:
            if item.cmd.execstr:
                if 'push(' in item.cmd.execstr:
                    titlenote = vd.options.disp_menu_push + ' '
                if 'input' in item.cmd.execstr:
                    title += vd.options.disp_menu_input

                if maxbinding:
                    revbinds = sheet.revbinds.get(item.cmd.longname, [])
                    if revbinds:
                        mainbinding = vd.prettykeys(revbinds[0])

        # actually display the menu item
        title += ' '*(w-dispwidth(pretitle)-dispwidth(item.title)+1) # padding

        menudraw(scr, y+i, x+1, pretitle+title, attr)
        if maxbinding and mainbinding:
            menudraw(scr, y+i, x+1+w-dispwidth(mainbinding), mainbinding, attr.update(colors.keystrokes))
        menudraw(scr, y+i, x+2+w, titlenote, attr)
        menudraw(scr, y+i, x+3+w, ls, colors.color_menu)

        vd.onMouse(scr, x, y+i, w+3, 1,
                BUTTON1_PRESSED=lambda y,x,key,p=sheet.activeMenuItems[:level]+[j]: sheet.pressMenu(*p),
                BUTTON2_PRESSED=vd.nop,
                BUTTON3_PRESSED=vd.nop,
                BUTTON1_CLICKED=lambda y,x,key,p=sheet.activeMenuItems[:level]+[j]: sheet.pressMenu(*p),
                BUTTON1_RELEASED=vd.nop,
                BUTTON2_RELEASED=vd.nop,
                BUTTON3_RELEASED=vd.nop)

        i += 1


@VisiData.api
def nop(vd, *args, **kwargs):
    return False


def _done(vd, *args, **kwargs):
    'Accept and execute current menu item (like pressing Enter).'
    return True


@BaseSheet.property
@drawcache
def menus(sheet):
    'List of hierarchical menu items for commands available on this sheet.'
    def _menus(sheet, item):
        if item.longname:
            cmd = sheet.getCommand(item.longname)
            if cmd:
                item.cmd = cmd
                if vd.commands[item.longname].get('TableSheet', None):
                    item.obj = 'TableSheet'
                elif vd.commands[item.longname].get('BaseSheet', None):
                    item.obj = 'BaseSheet'
                else:
                    item.obj = ''
                return item

        elif item.menus:
            menus = _menu_list(sheet, item.menus)
            if menus:
                title = getattr(item, 'title', '')
                return AttrDict(title=title, menus=menus, longname='')
        else:
            return item

    def _menu_list(sheet, menus):
        ret = []
        for i in menus:
            m = _menus(sheet, i)
            if m:
                ret.append(m)
        return ret

    return _menu_list(sheet, vd.menus)


@VisiData.api
def drawMenu(vd, scr, sheet):
    h, w = scr.getmaxyx()
    scr.addstr(0, 0, ' '*(w-1), colors.color_menu.attr)
    disp_menu_boxchars = sheet.options.disp_menu_boxchars
    x = 1
    ymax = 4
    toplevel = sheet.menus
    for i, item in enumerate(toplevel):
        if sheet.activeMenuItems and i == sheet.activeMenuItems[0]:
            cattr = colors.color_menu_active
            vd.drawSubmenu(scr, sheet, 1, x, item.menus, 1, disp_menu_boxchars)
        else:
            cattr = colors.color_menu

        menutitle = ' ' + ''.join(f'[:underline]{ch}[/]' if ch.isupper() else ch for ch in item.title) + ' '
        menudraw(scr, 0, x, menutitle, cattr)

        vd.onMouse(scr, x, 0, dispwidth(item.title)+2, 1,
                BUTTON1_PRESSED=lambda y,x,key,i=i,sheet=sheet: sheet.pressMenu(i),
                BUTTON2_PRESSED=vd.nop,
                BUTTON3_PRESSED=vd.nop,
                BUTTON1_CLICKED=lambda y,x,key,i=i,sheet=sheet: sheet.pressMenu(i),
                BUTTON1_RELEASED=vd.nop,
                BUTTON2_RELEASED=vd.nop,
                BUTTON3_RELEASED=vd.nop)
        x += dispwidth(item.title)+2

    rightdisp = sheet.options.disp_menu_fmt.format(sheet=sheet, vd=vd)
    menudraw(scr, 0, x+4, rightdisp, colors.color_menu)

    if not sheet.activeMenuItems:
        return

    currentItem = sheet.getMenuItem()
    cmd = currentItem.cmd

    if not cmd:
        return

    # helpbox
    sidelines = []
    if 'push(' in cmd.execstr:
        sidelines += [vd.options.disp_menu_push + ' pushes sheet']
    if 'input' in cmd.execstr:
        sidelines += [vd.options.disp_menu_input + ' needs input']

    helpattr = colors.color_menu_help
    helpx = 30
    helpw = min(w-helpx-4, 76)
    ls,rs,ts,bs,tl,tr,bl,br,lsr,rsl = disp_menu_boxchars
    helplines = textwrap.wrap(cmd.helpstr or '(no help available)', width=helpw-4)

    # place helpbox just below deepest menu
    menuh = 2+sum(sheet.activeMenuItems[1:-1])
    menuh += len(sheet.getMenuItem(sheet.activeMenuItems[:-1]).menus)
    menuy = 16 # min(menuh, h-len(helplines)-3)

    y = menuy
    menudraw(scr, y, helpx, tl+ts*(helpw-2)+tr, helpattr) # top line
    y += 1

    # cmd.helpstr text
    for i, line in enumerate(helplines):
        menudraw(scr, y+i, helpx, ls+' '+line+' '*(helpw-dispwidth(line)-3)+rs, helpattr)
    y += len(helplines)

    if sidelines:
        menudraw(scr, y, helpx, ls+' '*(helpw-2)+rs, helpattr)
        for i, line in enumerate(sidelines):
            menudraw(scr, y+i+1, helpx, ls+'    '+line+' '*(helpw-dispwidth(line)-6)+rs, helpattr)
        y += len(sidelines)+1

    menudraw(scr, y, helpx, bl+bs*(helpw-2)+br, helpattr)

    mainbinding = sheet.revbinds.get(cmd.longname, [None])[0]
    if mainbinding:
        menudraw(scr, menuy, helpx+2, rsl, helpattr)
        ks = vd.prettykeys(mainbinding or '(unbound)')
        menudraw(scr, menuy, helpx+3, ' '+ks+' ', colors.color_menu_active)
        menudraw(scr, menuy, helpx+2+dispwidth(ks)+3, lsr, helpattr)
    menudraw(scr, menuy, helpx+19, ' '+cmd.longname+' ', helpattr)

    vd.onMouse(scr, helpx, menuy, helpw, y-menuy+1,
               BUTTON1_PRESSED=_done,
               BUTTON1_CLICKED=_done,
               BUTTON1_RELEASED=vd.nop,
               BUTTON2_RELEASED=vd.nop,
               BUTTON3_RELEASED=vd.nop)


@BaseSheet.api
def pressMenu(sheet, *args):
    '''Navigate to given menupath in *args* and activate menu if not already activated. Return True if pressing current item.
    Example: sheet.pressMenu("Help", "Version")
    '''
    ret = False
    p = _intMenuPath(sheet, args)
    if p == sheet.activeMenuItems:  # clicking on current menu item
        if sheet.getMenuItem(p).longname:
            ret = True
        else:
            p += [0]

    sheet.activeMenuItems = p

    if not vd.menuRunning:
        vd.runMenu()

    return ret


@BaseSheet.api
def checkMenu(sheet):
    sheet.activeMenuItems[0] %= len(sheet.menus)


@VisiData.api
def runMenu(vd):
    'Activate menu, with sheet.activeMenuItems containing the navigated menu path.  Does not return until menu is deactivated.'
    old_disp_menu = vd.options.disp_menu

    vd.options.disp_menu=True
    vd.menuRunning = True
    sheet = vd.activeSheet
    vd.setWindows(vd.scrFull)
    nEscapes = 0

    def _clickedDuringMenu():
        r = vd.parseMouse(menu=vd.scrMenu, top=vd.winTop, bot=vd.winBottom)
        f = vd.getMouse(r.x, r.y, r.keystroke)
        if f:
            if f(r.y, r.x, r.keystroke):  # call each function until one returns a true-ish value
                return 'doit'
        else:
            return 'offmenu'

    try:
      while True:
        if len(sheet.activeMenuItems) < 2:
            sheet.activeMenuItems.append(0)

        vd.draw_all()

        k = vd.getkeystroke(vd.scrMenu, sheet)

        if not k:
            continue

        currentItem = sheet.getMenuItem()

        if k == '^[':  # ESC
            nEscapes += 1  #1470
            if nEscapes > 1:
                return
            continue
        else:
            nEscapes = 0

        if k in ['^C', '^Q', 'q', '^H', 'KEY_BACKSPACE']:
            return

        elif k in ['KEY_MOUSE']:
            r = _clickedDuringMenu()
            if r == 'offmenu':
                return  # clicking off the menu is an escape
            elif r == 'doit':
                break

        elif k in ['KEY_RIGHT', 'l']:
            if currentItem.menus and sheet.activeMenuItems[1] != 0:  # not first item
                sheet.activeMenuItems.append(0)
            else:
                sheet.activeMenuItems = [sheet.activeMenuItems[0]+1, 0]

        elif k in ['KEY_LEFT', 'h']:
            if len(sheet.activeMenuItems) > 2:
                sheet.activeMenuItems.pop(-1)
            else:
                sheet.activeMenuItems = [sheet.activeMenuItems[0]-1, 0]

        elif k in ['KEY_DOWN', 'j']:
            sheet.activeMenuItems[-1] += 1

        elif k in ['KEY_UP', 'k']:
            sheet.activeMenuItems[-1] -= 1

        elif k in [ENTER, ' ', '^J', '^M']:
            if currentItem.menus:
                sheet.activeMenuItems.append(0)
            else:
                break

        elif k in main_menu.keys():
            sheet.pressMenu(main_menu[k])

        else:
            vd.warning(f'unknown keystroke {k}')

        sheet.checkMenu()

    finally:
        vd.menuRunning = False
        sheet.activeMenuItems = []
        vd.options.disp_menu=old_disp_menu

    vd.draw_all()
    sheet.execCommand(currentItem.longname)

main_menu = {'f': 'File', 'e': 'Edit', 'v': 'View', 'c': 'Column', 'r': 'Row', 'd': 'Data', 'p': 'Plot', 's': 'System', 'h': 'Help'}

BaseSheet.addCommand('Alt+f', 'menu-file', 'pressMenu("File")', '')
BaseSheet.addCommand('Alt+e', 'menu-edit', 'pressMenu("Edit")', '')
BaseSheet.addCommand('Alt+v', 'menu-view', 'pressMenu("View")', '')
BaseSheet.addCommand('Alt+c', 'menu-column', 'pressMenu("Column")', '')
BaseSheet.addCommand('Alt+r', 'menu-row', 'pressMenu("Row")', '')
BaseSheet.addCommand('Alt+d', 'menu-data', 'pressMenu("Data")', '')
BaseSheet.addCommand('Alt+p', 'menu-plot', 'pressMenu("Plot")', '')
BaseSheet.addCommand('Alt+s', 'menu-system', 'pressMenu("System")', '')
BaseSheet.addCommand('Alt+h', 'menu-help', 'pressMenu("Help")', 'open the Help menu')
BaseSheet.bindkey('Ctrl+H', 'menu-help')
BaseSheet.bindkey('KEY_BACKSPACE', 'menu-help')

vd.addGlobals({'Menu': Menu})



import collections
import os
from copy import copy

from visidata import vd
from visidata import Sheet, BaseSheet, VisiData, IndexSheet, Path, Progress, TypedExceptionWrapper, TypedWrapper, UNLOADED

vd.option('safe_error', '#ERR', 'error string to use while saving', replay=True)
vd.option('save_encoding', 'utf-8', 'encoding passed to codecs.open when saving a file', replay=True, help=vd.help_encoding)

@Sheet.api
def safe_trdict(vs):
    'returns string.translate dictionary for replacing tabs and newlines'
    if vs.options.safety_first:
        delim = vs.options.delimiter
        trdict = {
             0: '', #  strip NUL completely
            10: vs.options.tsv_safe_newline,  # \n
            13: vs.options.tsv_safe_newline,  # \r
        }
        if not delim or ord(delim) in trdict:
            vd.fail(f'cannot use delimiter {repr(delim)} with safety_first')
        trdict[ord(delim)] = vs.options.tsv_safe_tab # \t
        return trdict
    return {}


@Sheet.api
def iterdispvals(sheet, *cols, format=False):
    'For each row in sheet, yield OrderedDict of values for given cols.  Values are typed if format=False, or a formatted display string if format=True.'
    if not cols:
        cols = sheet.visibleCols

    transformers = collections.OrderedDict()  # list of transformers for each column in order
    trdict = sheet.safe_trdict()
    for col in cols:
        transformers[col] = [ col.type ]
        if format:
            formatMaker = getattr(col, 'formatter_'+(col.formatter or sheet.options.disp_formatter))
            transformers[col].append(formatMaker(col._formatdict))
        if trdict:
            transformers[col].append(lambda v,trdict=trdict: v.translate(trdict))

    options_safe_error = sheet.options.safe_error
    for r in Progress(sheet.rows):
        dispvals = collections.OrderedDict()  # [col] -> value
        for col, transforms in transformers.items():
            try:
                dispval = col.getValue(r)

            except Exception as e:
                dispval = options_safe_error or str(e)

            try:
                for t in transforms:
                    if dispval is None:
                        break
                    elif isinstance(dispval, TypedExceptionWrapper):
                        dispval = options_safe_error or str(dispval)
                        break
                    elif isinstance(dispval, TypedWrapper):
                        dispval = ''
                        break
                    else:
                        dispval = t(dispval)

                if dispval is None and format:
                    dispval = ''
            except Exception as e:
                dispval = str(dispval)

            dispvals[col] = dispval

        yield dispvals


@Sheet.api
def itervals(sheet, *cols, format=False):
    for row in sheet.iterdispvals(*cols, format=format):
        yield [row[c] for c in cols]

@BaseSheet.api
def getDefaultSaveName(sheet):
    src = getattr(sheet, 'source', None)
    if hasattr(src, 'scheme') and src.scheme:
        return src.name + src.suffix
    if isinstance(src, Path):
        if src.given == '-':
            return f'stdin.{sheet.options.save_filetype}'
        if sheet.options.is_set('save_filetype', sheet):
            # if save_filetype is over-ridden from default, use it as the extension
            return str(src.with_suffix('')) + '.' + sheet.options.save_filetype
        return str(src)
    else:
        return sheet.name+'.'+getattr(sheet, 'filetype', sheet.options.save_filetype)


@VisiData.api
def saveCols(vd, cols):
    sheet = cols[0].sheet
    vs = copy(sheet)
    vs.columns = list(cols)
    vs.rows = sheet.rows
    if len(cols) == 1:
        savedcoltxt = cols[0].name + ' column'
    else:
        savedcoltxt = '%s columns' % len(cols)
    path = vd.inputPath('save %s to: ' % savedcoltxt, value=vs.getDefaultSaveName())
    vd.saveSheets(path, vs)


@VisiData.api
def saveSheets(vd, givenpath, *vsheets, confirm_overwrite=True):
    '''Save all *vsheets* to *givenpath*. Async.
    Callers should be careful not to call reload() while saveSheets is still running.
    Use vd.sync(saveSheets) to wait for the save to finish.'''

    if not vsheets: # blank tuple
        vd.warning('no sheets to save')
        return
    if not givenpath.name:
        vd.warning('no save path given')
        return

    unloaded = [ vs for vs in vsheets if vs.rows is UNLOADED ]
    vd.sync(*vd.ensureLoaded(unloaded))

    filetypes = [givenpath.ext.lower(), vd.options.save_filetype.lower()]

    vd.clearCaches()

    for ft in filetypes:
        savefunc = getattr(vsheets[0], 'save_' + ft, None) or getattr(vd, 'save_' + ft, None)
        if savefunc:
            filetype = ft
            break

    if savefunc is None:
        vd.fail(f'no function to save as {", ".join(filetypes)}')

    if confirm_overwrite:
        vd.confirmOverwrite(givenpath)

    vd.status('saving %s sheets to %s as %s' % (len(vsheets), givenpath.given, filetype))

    if not givenpath.given.endswith('/'):  # forcibly specify save individual files into directory by ending path with /
        for vs in vsheets:
            vs.hasBeenModified = False
        # savefuncs(vd, p, *vsheets) will have 2 argcount (*vsheets does not get counted as an arg)
        # savefuncs(vd, p, vs) will have 3 argcount (vs counts as an arg, along with vd, path)
        if savefunc.__code__.co_argcount == 3 and len(vsheets) > 1:
            vd.fail(f'cannot save multiple {filetype} sheets to non-dir')
        return vd.execAsync(savefunc, givenpath, *vsheets)

    # path is a dir

    # save as individual files in the givenpath directory
    try:
        os.makedirs(givenpath, exist_ok=True)
    except FileExistsError:
        pass

    if not givenpath.is_dir():
        vd.fail(f'cannot save multiple {filetype} sheets to non-dir')

    def _savefiles(vsheets, givenpath, savefunc, filetype):
        for vs in vsheets:
            p = Path((givenpath / vs.name).with_suffix('.'+filetype))
            savefunc(p, vs)
            vs.hasBeenModified = False

        vd.status(f'{givenpath} save finished')  #2157

    return vd.execAsync(_savefiles, vsheets, givenpath, savefunc, filetype)


@VisiData.api
def save_zip(vd, p, *vsheets):
    vd.clearCaches()

    import tempfile
    import zipfile
    with tempfile.TemporaryDirectory() as tmpdir:
        with zipfile.ZipFile(str(p), 'w', zipfile.ZIP_DEFLATED, allowZip64=True, compresslevel=9) as zfp:
            for vs in Progress(vsheets):
                filetype = vs.options.save_filetype
                tmpp = Path(f'{tmpdir}{vs.name}.{filetype}')
                savefunc = getattr(vs, 'save_' + filetype, None) or getattr(vd, 'save_' + filetype, None)
                savefunc(tmpp, vs)
                zfp.write(tmpp, f'{vs.name}.{vs.options.save_filetype}')


@VisiData.api
def save_txt(vd, p, *vsheets):
    if len(vsheets) == 1 and vsheets[0].nVisibleCols > 1:  #2173
        return vd.save_tsv(p, vsheets[0])

    with p.open(mode='w', encoding=vsheets[0].options.save_encoding) as fp:
        for vs in vsheets:
            unitsep = vs.options.delimiter
            rowsep = vs.options.row_delimiter
            for dispvals in vs.iterdispvals(*vs.visibleCols, format=True):
                fp.write(unitsep.join(dispvals.values()))
                fp.write(rowsep)


@BaseSheet.api
def rootSheet(sheet):
    r = sheet
    while isinstance(r.source, BaseSheet):
        r = r.source

    return r


BaseSheet.addCommand('^S', 'save-sheet', 'vd.saveSheets(inputPath("save to: ", value=getDefaultSaveName()), sheet)', 'save current sheet to filename in format determined by extension (default .tsv)')
BaseSheet.addCommand('', 'save-sheet-really', 'vd.saveSheets(Path(getDefaultSaveName()), sheet, confirm_overwrite=False)', 'save current sheet without asking for filename or confirmation')
BaseSheet.addCommand('', 'save-source', 'vd.saveSheets(rootSheet().source, rootSheet())', 'save root sheet to its source')
BaseSheet.addCommand('g^S', 'save-all', 'vd.saveSheets(inputPath("save all sheets to: "), *vd.stackedSheets)', 'save all sheets to given file or directory)')
IndexSheet.addCommand('g^S', 'save-selected', 'vd.saveSheets(inputPath("save %d sheets to: " % nSelectedRows, value="_".join(getattr(vs, "name", None) or "blank" for vs in selectedRows)), *selectedRows)', 'save all selected sheets to given file or directory')
Sheet.addCommand('', 'save-col', 'saveCols([cursorCol])', 'save current column only to filename in format determined by extension (default .tsv)')
Sheet.addCommand('', 'save-col-keys', 'saveCols(keyCols + [cursorCol])', 'save key columns and current column to filename in format determined by extension (default .tsv)')

vd.addMenuItems('''
    File > Save > current sheet > save-sheet
    File > Save > all sheets > save-all
    File > Save > current column > save-col
    File > Save > keys and current column > save-col-keys
''')



from visidata import (
    ColumnItem,
    PythonSheet,
    VisiData,
    asyncthread,
    deduceType,
    vd,
)


@VisiData.api
def open_toml(vd, p):
    return TomlSheet(p.base_stem, source=p)


class TomlSheet(PythonSheet):
    """A Sheet representing the top level of a loaded TOML file.

    This is an intentionally minimal loader with cues taken from
    VisiData built-in JSON and Python object sheet types.
    """
    guide = '''# Toml Sheet
This sheet represents the top level of {sheet.source.name}.{sheet.source.ext}.

Each cell within this sheet can contain dictionaries (representing TOML key:value pairs), lists (representing TOML arrays), or scalars.

Some helpful commands when working with cells of lists and dictionaries:

- `(` (`expand-col`) on a column with lists or dictionaries will "expand" the structures in the cells into new columns within the current sheet.
- `zEnter` on a cell with lists or dictionaries will "dive" into the current cell, expanding its structures into rows and columns in a separate sheet.
'''

    rowtype = "values"  # rowdef: dict values, possibly nested

    def loader(self):
        """Loading a TOML file produces a single dict. Use
        its keys as column headings, and populate a single
        row.
        """
        self.columns = []
        self.rows = []

        try:
            # Python 3.11+
            import tomllib
        except ModuleNotFoundError:
            # Python 3.10 and below
            tomllib = vd.importExternal("tomli")

        data = tomllib.loads(self.source.read_text())
        for k, v in data.items():
            self.addColumn(ColumnItem(k, type=deduceType(v)))
        self.addRow(data)


vd.addGlobals(
    {
        "TomlSheet": TomlSheet,
    }
)



import re
import os

from visidata import vd, date, asyncthread, VisiData, Progress, Sheet, Column, ItemColumn, deduceType, TypedWrapper, setitem, AttrDict


vd.option('airtable_auth_token', '', 'Airtable API key from https://airtable.com/account')

airtable_regex = r'^https://airtable.com/(app[A-Za-z0-9]+)/(tbl[A-Za-z0-9]+)/?(viw[A-z0-9]+)?'

@VisiData.api
def guessurl_airtable(vd, p, response):
    m = re.search(airtable_regex, p.given)
    if m:
        return dict(filetype='airtable', _likelihood=10)


@VisiData.api
def open_airtable(vd, p):
    pyairtable = vd.importExternal('pyairtable')

    token = os.environ.get('AIRTABLE_AUTH_TOKEN') or vd.options.airtable_auth_token
    if not token:
        vd.requireOptions('airtable_auth_token', help='https://support.airtable.com/docs/creating-and-using-api-keys-and-access-tokens')

    m = re.search(airtable_regex, p.given)
    if not m:
        vd.fail('invalid airtable url')

    app, tbl, viw = m.groups()
    return AirtableSheet('airtable', source=p,
                         airtable_auth_token=token,
                         airtable_base=app,
                         airtable_table=tbl,
                         airtable_view=viw)


class AirtableSheet(Sheet):
    guide = '''
        # Airtable
        This sheet is a read-only download of all records in a table at _airtable.com_.
    '''
    rowtype = 'records'  # rowdef: dict

    columns = [
        ItemColumn('id', 'id', type=str, width=0),
        ItemColumn('createdTime', 'createdTime', type=date, width=0)
    ]

    def iterload(self):
        self.fields = set()

        for page in self.api.iterate(self.airtable_base, self.airtable_table, view=self.airtable_view):
            for row in page:
                yield row

                for field, value in row['fields'].items():
                    if field not in self.fields:
                        col = ItemColumn('fields.'+field, type=deduceType(value))
                        self.addColumn(col)
                        self.fields.add(field)

    def newRow(self):
        return AttrDict(fields=AttrDict())


@AirtableSheet.lazy_property
def api(self):
    import pyairtable
    return pyairtable.Api(self.airtable_auth_token)



"""
Filename: vdfec.py
Last updated: 2019-04-21
Home: https://github.com/jsvine/visidata-plugins
Author: Jeremy Singer-Vine

# Installation

- Install fecfile: `pip install fecfile`
- Add vdfec.py to your ~/.visidata directory
- Add "import vdfec" to your ~/.visidatarc file

# Usage

vdfec.py enables VisiData to load .fec files from the Federal Election Commission.

Once saved to your ~/.visidata directory, and imported via your ~/.visidatarc file,
you should be able to interactively explore .fec files as you would any other 
filetype in VisiData.

From the command line:

    vd path/to/my/file.fec

# Thanks

vdfec.py depends heavily on Evan Sonderegger's `fecfile` Python library: https://esonderegger.github.io/fecfile/

... which in turn is based on Derek Willis's `Fech` Ruby library: https://github.com/dwillis/Fech

... and Chris Zubak-Skees' transformation of `Fech`'s form-and-field mappings: https://github.com/PublicI/fec-parse/blob/master/lib/renderedmaps.js

Thanks to all who have contributed to those projects.

"""

from copy import copy
from visidata import (
    vd,
    VisiData,
    Path,
    Sheet,
    TextSheet,
    ColumnAttr,
    ColumnItem,
    asyncthread,
    Progress,
)

class DiveSheet(Sheet):
    "A deeply-diveable, quick-diving sheet."

    def reload(self):
        mapping = self.source

        self.columns = []
        self.rows = []

        self.key_type = str
        self.size = len(mapping)

        if self.size == 0:
            return

        if isinstance(mapping, list):
            first = mapping[0]
            if isinstance(first, dict):
                colgetter = lambda x: x.keys()
            elif isinstance(first, list):
                colgetter = lambda x: list(range(len(x)))
            else:
                mapping = dict(enumerate(mapping))
                self.key_type = int
                self.size = len(mapping)

        if isinstance(mapping, dict):
            self.is_keyvalue = True
            if self.size:
                max_key_len = max(map(len, map(str, mapping.keys())))
                key_width = min(50, max(max_key_len + 2, 6))
            else:
                key_width = None

            self.addColumn(ColumnItem(
                "key",
                width = key_width,
                type = self.key_type
            ))
            self.addColumn(ColumnItem("value"))
            self.setKeys(self.columns[:1])

            for k, v in mapping.items():
                self.addRow({ "key": k, "value": v })

        elif isinstance(mapping, list):
            self.is_keyvalue = False
            indices = [] 
            for item in mapping:
                try:
                    cols = colgetter(item)
                    for col in cols:
                        if col not in indices:
                            self.addColumn(ColumnItem(col))
                            indices.append(col)

                    self.addRow(item)

                except Exception as e:
                    vd.warning("Can't dive on lists with heterogeneous item types.")
                    return False

    def openRow(self, row):
        if self.is_keyvalue:
            cell = row["value"]
            if isinstance(cell, (list, dict)):
                vs = self.__class__(self.name, row["key"], source = cell)
            else:
                vd.warning("Nothing to dive into.")
                return
        else:
            vs = self.__class__(self.name, "row", source = self.row)

        success = vs.reload()
        if success == False:
            vd.fail('could not reload new sheet')
        return vs


class FECItemizationSheet(Sheet):
    "A sheet to display a list of FEC itemizations from a given form/schedule."

    rowtype = "itemizations"

    @asyncthread
    def reload(self):
        self.rows = []
        self.columns = []

        if len(self.source) == 0:
            return

        for i, row in enumerate(Progress(self.source, total = len(self.source))):
            if i == 0:
                self.set_columns_from_row(row)
            self.addRow(row)
    
    def set_columns_from_row(self, row):
        self.columns.clear()
        for i, name in enumerate(row.keys()):
            self.addColumn(ColumnItem(name))

    def openRow(self, row):
        return row

class FECScheduleSheet(Sheet):
    "A sheet to display the list of itemized schedules in a filing."

    rowtype = "schedules"

    columns = [
        ColumnAttr("schedule", "schedule_name", width = 14),
        ColumnAttr("name", width = 0),
        ColumnAttr("size", type = int),
    ]

    nKeys = 1

    @asyncthread
    def reload(self):
        self.rows = []

        for schedule_name in self.source.keys():
            vs = FECItemizationSheet(
                self.name, schedule_name,
                schedule_name = schedule_name,
                source = self.source[schedule_name],
                size = len(self.source[schedule_name]),
            )
            self.addRow(vs)

    def openRow(self, row):
        return row

COMPONENT_SHEET_CLASSES = {
    "header": DiveSheet,
    "summary": DiveSheet,
    "itemization": FECScheduleSheet,
    "text": FECItemizationSheet,
    "F99_text": TextSheet,
}

class FECFiling(Sheet):
    "A sheet representing an entire .fec file."

    rowtype = "components"
    filing = None

    columns = [
        ColumnAttr("component", "component_name", width = 14),
        ColumnAttr("name", width = 0),
        ColumnAttr("size", type = int),
    ]

    nKeys = 1

    @asyncthread
    def reload(self):
        from fecfile import fecparser
        self.rows = []  # rowdef:  Sheet, of a type from COMPONENT_SHEET_CLASSES.values()

        row_dict = { }
        itemization_subsheets = {}

        def addSheetRow(component_name):
            "On first encountering a component, add a row to the filing sheet"

            cls = COMPONENT_SHEET_CLASSES[component_name]

            source_cls = list if cls in [
                    FECItemizationSheet,
                    TextSheet
                ] else dict

            vs = cls(
                self.name, component_name,
                component_name = component_name,
                source = source_cls(),
                size = 0,
            )

            vs.reload()
            row_dict[component_name] = vs
            self.addRow(vs)

        src = Path(self.source.resolve())

        item_iter = fecparser.iter_lines(src, { "as_strings": True })

        for item in item_iter:
            dtype = item.data_type 
            if dtype not in row_dict.keys():
                addSheetRow(dtype)

            sheet_row = row_dict[dtype]

            if dtype in [ "header", "summary" ]:
                sheet_row.source = item.data
                sheet_row.reload()

            elif dtype == "text":
                if len(sheet_row.source) == 0:
                    sheet_row.set_columns_from_row(item.data)
                sheet_row.source.append(item.data)
                sheet_row.addRow(item.data)
                sheet_row.size += 1

            elif dtype == "F99_text":
                sheet_row.source = item.data.split("\n")
                sheet_row.size = len(sheet_row.source)

            elif dtype == "itemization":
                form_type = item.data["form_type"]

                if form_type[0] == "S":
                    form_type = "Schedule " + item.data["form_type"][1]

                if form_type not in sheet_row.source:
                    sheet_row.source[form_type] = [ ] 
                    subsheet = FECItemizationSheet(
                        sheet_row.name, form_type,
                        schedule_name = form_type,
                        source = [ ],
                        size = 0,
                    )
                    subsheet.reload()
                    subsheet.set_columns_from_row(item.data)
                    sheet_row.addRow(subsheet)
                    itemization_subsheets[form_type] = subsheet
                else:
                    subsheet = itemization_subsheets[form_type]

                subsheet.addRow(item.data)
                subsheet.source.append(item.data)
                subsheet.size += 1

                sheet_row.source[form_type].append(item.data)
                sheet_row.size += 1

    def openRow(self, row):
        return row

@VisiData.api
def open_fec(vd, p):
    return FECFiling(p.base_stem, source=p)

vd.addGlobals({
    "DiveSheet": DiveSheet
})



'''
# vgit: VisiData wrapper for git

The syntax for vgit is the same as the syntax for git.
By default, will pass the command to git verbatim, as quickly as possible.
If vgit can provide an interactive interface for a particular subcommand,
it will open the sheet returned by vd.git_<subcommand>(path, args).
'''

import os
import sys


def vgit_cli():
    import visidata
    from visidata import vd, Path

    args = sys.argv[1:]
    flDebug = '--debug' in args
    if flDebug:
        args.remove('--debug')

    if not args:
        args = ['help']

    func = getattr(vd, 'git_'+args[0], None)
    if func:
        vd.loadConfigAndPlugins()
        vd.status(visidata.__version_info__)
        vd.domotd()
        if flDebug:
            vd.options.debug = True

        rc = 0
        try:
            p = Path('.')
            vs = func(p, args[1:])
            if vs:
                vd.run(vs)
        except BrokenPipeError:
            os.dup2(os.open(os.devnull, os.O_WRONLY), sys.stdout.fileno()) # handle broken pipe gracefully
        except visidata.ExpectedException as e:
            print(str(e))
        except Exception as e:
            rc = 1
            vd.exceptionCaught(e)
            if flDebug:
                raise

        sys.stderr.flush()
        sys.stdout.flush()
        os._exit(rc)  # cleanup can be expensive

    import subprocess
    return subprocess.run(['git', *args]).returncode



"""
Column types and utility commands related to IP addresses.
"""
from ipaddress import ip_address, ip_network, _BaseNetwork

from visidata import vd
from visidata.sheets import Column, TableSheet


vd.addType(ip_address, icon=":", formatter=lambda fmt, ip: str(ip), name='ip_address')
vd.addType(ip_network, icon="/", formatter=lambda fmt, ip: str(ip), name='ip_network')


def isSupernet(cell, network, isNull):
    """Is `cell` a supernet of `network`?

    Treat nulls as false, and perform conversions to IP network objects only
    if necessary.
    """
    if isNull(cell):
        return False
    if not isinstance(cell, _BaseNetwork):
        try:
            cell = ip_network(str(cell).strip())
        except ValueError:
            return False
    return cell.supernet_of(network)


@Column.api
def selectSupernets(col, ip):
    """Select rows based on network containment

    Given an IP address (e.g. 10.0.0.0) or network (e.g. 10.0.0.0/8) as input,
    select rows whose network address space completely contains the input network.
    """
    if not ip:
        return

    sheet = col.sheet
    network = ip_network(ip.strip())
    isNull = sheet.isNullFunc()

    vd.status(f'selecting rows where {col.name} is a supernet of "{str(network)}"')
    sheet.select(
        [
            row
            for row in sheet.rows
            if isSupernet(col.getTypedValue(row), network, isNull)
        ]
    )


TableSheet.addCommand(
    None,
    "type-ipaddr",
    "cursorCol.type=ip_address",
    "set type of current column to IP address",
)
TableSheet.addCommand(
    None,
    "type-ipnet",
    "cursorCol.type=ip_network",
    "set type of current column to IP network",
)
TableSheet.addCommand(
    None,
    "select-supernets",
    'cursorCol.selectSupernets(input("ip or cidr block: "))',
    "select rows where the CIDR block value includes the input address space",
)



"""
# Usage

Duplicates are determined by the sheet's key columns.

If no key columns are specified, then a duplicate row is one where the values
of *all non-hidden* columns are exactly the same as a row that occurs earlier
in the sheet.

If key columns *are* specified, then duplicates are detected based on the
values in just those columns.

## Commands

- `select-duplicate-rows` sets the selection status in VisiData to `selected`
  for each row in the active sheet that is a duplicate of a prior row.

- `dedupe-rows` pushes a new sheet in which only non-duplicate rows in the
  active sheet are included.
"""


__author__ = "Jeremy Singer-Vine <jsvine@gmail.com>"

from copy import copy

from visidata import Sheet, TableSheet, asyncthread, Progress, vd


def gen_identify_duplicates(sheet):
    """
    Takes a sheet, and returns a generator yielding a tuple for each row
    encountered. The tuple's structure is `(row_object, is_dupe)`, where
    is_dupe is True/False.

    See note in Usage section above regarding how duplicates are determined.
    """

    keyCols = sheet.keyCols

    cols_to_check = None
    if len(keyCols) == 0:
        vd.warning("No key cols specified. Using all columns.")
        cols_to_check = sheet.visibleCols
    else:
        cols_to_check = sheet.keyCols

    seen = set()
    for r in sheet.rows:
        vals = tuple(col.getValue(r) for col in cols_to_check)
        is_dupe = vals in seen
        if not is_dupe:
            seen.add(vals)
        yield (r, is_dupe)


@Sheet.api
@asyncthread
def select_duplicate_rows(sheet, duplicates=True):
    """
    Given a sheet, sets the selection status in VisiData to `selected` for each
    row that is a duplicate of a prior row.

    If `duplicates = False`, then the behavior is reversed; sets the selection
    status to `selected` for each row that is *not* a duplicate.
    """
    before = len(sheet.selectedRows)

    gen = gen_identify_duplicates(sheet)
    prog = Progress(gen, gerund="selecting", total=sheet.nRows)

    for row, is_dupe in prog:
        if is_dupe == duplicates:
            sheet.selectRow(row)

    sel_count = len(sheet.selectedRows) - before

    more_str = " more" if before > 0 else ""

    vd.status(f"selected {sel_count}{more_str} {sheet.rowtype}")


@Sheet.api
def dedupe_rows(sheet):
    """
    Given a sheet, pushes a new sheet in which only non-duplicate rows are
    included.
    """
    vs = copy(sheet)
    vs.name += "_deduped"

    @asyncthread
    def _reload(self=vs):
        self.rows = []
        gen = gen_identify_duplicates(sheet)
        prog = Progress(gen, gerund="deduplicating", total=sheet.nRows)
        for row, is_dupe in prog:
            if not is_dupe:
                self.addRow(row)

    vs.reload = _reload
    return vs


# Add longname-commands to VisiData to execute these methods
TableSheet.addCommand(None, "select-duplicate-rows", "sheet.select_duplicate_rows()", "select each row that is a duplicate of a prior row")
TableSheet.addCommand(None, "dedupe-rows", "vd.push(sheet.dedupe_rows())", "open new sheet in which only non-duplicate rows in the active sheet are included")

vd.addMenuItems('''
    Row > Select > duplicate rows > select-duplicate-rows
    Data > Deduplicate rows > dedupe-rows
''')

"""
# Changelog

## 0.2.0 - 2021-09-22

Use `vd.warning(...)` instead of `warning(...)`

## 0.1.0 - 2020-10-09

Revised for compatibility with VisiData 2.x

## 0.0.1 - 2019-01-01

Internal change, no external effects: Migrates from ._selectedRows to .selectedRows.

## 0.0.0 - 2018-12-30

Initial release.
"""



import collections

#1179  Line Separated Values for e.g. awk

from visidata import VisiData, Sheet, ItemColumn


@VisiData.api
def open_lsv(vd, p):
    return LsvSheet(p.base_stem, source=p)


@VisiData.api
def save_lsv(vd, p, *vsheets):
    vs = vsheets[0]
    with p.open(mode='w', encoding=vs.options.save_encoding) as fp:
        for row in vs.iterrows():
            for col in vs.visibleCols:
                fp.write('%s: %s\n' % (col.name, col.getDisplayValue(row)))
            fp.write('\n')


class LsvSheet(Sheet):
    def addRow(self, row, **kwargs):
        super().addRow(row, **kwargs)
        for k in row:
            if k not in self._knownCols:
                self.addColumn(ItemColumn(k))
                self._knownCols.add(k)


    def iterload(self):
        self.columns = []
        self.rows = []
        self._knownCols = set()
        row = collections.defaultdict(str)
        k = ''

        with self.open_text_source() as fp:
            for line in fp:
                line = line.strip()
                if not line:
                    yield row
                    row = collections.defaultdict(str)

                if ':' in line:
                    k, line = line.split(':', maxsplit=1)
                # else append to previous k

                row[k.strip()] += line.strip()

        if row:
            yield row



import itertools

from visidata import Sheet, ListAggregator, SettableColumn
from visidata import vd, anytype, asyncthread, Progress

class RankAggregator(ListAggregator):
    '''
    Ranks start at 1, and each group's rank is 1 higher than the previous group.
    When elements are tied in ranking, each of them gets the same rank.
    '''
    def aggregate(self, col, rows) -> [int]:
        return self.aggregate_list(col, rows)

    def aggregate_list(self, col, rows) -> [int]:
        if not col.sheet.keyCols:
            vd.error('ranking requires one or more key columns')
        return self.rank(col, rows)

    def rank(self, col, rows):
        if col.keycol:
            vd.warning('rank aggregator is uninformative for key columns')
        def _key_progress(prog):
            def identity(val):
                prog.addProgress(1)
                return val
            return identity
        with Progress(gerund='ranking', total=4*col.sheet.nRows) as prog:
            p = _key_progress(prog) # increment progress every time p() is called
            # compile row data, for each row a list of tuples: (group_key, rank_key, rownum)
            rowdata = [(col.sheet.rowkey(r), col.getTypedValue(r), p(rownum)) for rownum, r in enumerate(rows)]
            # sort by row key and column value to prepare for grouping
            # If the column is in descending order, use descending order for within-group ranking.
            reverse = next((r for (c, r) in col.sheet.ordering if c == col or c == col.name), False)
            try:
                rowdata.sort(reverse=reverse, key=p)
                if reverse:
                    vd.status('ranking {col.name} in descending order')
            except TypeError as e:
                vd.fail(f'elements in a ranking column must be comparable: {e.args[0]}')
            rowvals = []
            #group by row key
            for _, group in itertools.groupby(rowdata, key=lambda v: v[0]):
                # within a group, the rows have already been sorted by col_val
                group = list(group)
                # rank each group individually
                group_ranks = rank_sorted_iterable([p(col_val) for _, col_val, rownum in group])
                rowvals += [(rownum, rank) for (_, _, rownum), rank in zip(group, group_ranks)]
            # sort by unique rownum, to make rank results match the original row order
            rowvals.sort(key=p)
            rowvals = [ rank for rownum, rank in rowvals ]
        return rowvals

vd.aggregators['rank'] = RankAggregator('rank', anytype, helpstr='list of ranks, when grouping by key columns', listtype=int)

def rank_sorted_iterable(vals_sorted) -> [int]:
    '''*vals_sorted* is an iterable whose elements form one or more groups.
    The iterable must already be sorted.'''

    ranks = []
    val_groups = itertools.groupby(vals_sorted)
    for rank, (_, val_group) in enumerate(val_groups, 1):
        for _ in val_group:
            ranks.append(rank)
    return ranks

@Sheet.api
@asyncthread
def addcol_sheetrank(sheet, rows):
    '''
    Each row is ranked within its sheet. Rows are ordered by the
    value of their key columns.
    '''
    if not sheet.keyCols:
        vd.error('ranking requires one or more key columns')
    colname = f'{sheet.name}_sheetrank'
    c = SettableColumn(name=colname, type=int)
    sheet.addColumnAtCursor(c)
    def _key_progress(prog):
        def identity(val):
            prog.addProgress(1)
            return val
        return identity
    with Progress(gerund='ranking', total=5*sheet.nRows) as prog:
        p = _key_progress(prog) # increment progress every time p() is called
        ordering = [(col, reverse) for (col, reverse) in sheet.ordering if col.keycol]
        rowkeys = [(sheet.rowkey(r), p(rownum), r) for rownum, r in enumerate(rows)]
        if ordering:
            vd.status('using custom ordering for keycol sort')
            keycols_ordered = [col for (col, reverse) in ordering]
            keycols_unordered = [keycol for keycol in sheet.keyCols if not keycol in keycols_ordered]
            ordering += [(keycol, False) for keycol in keycols_unordered]
            def _sortkey(e): # sort the rows by using the column
                p(None)
                return sheet.sortkey(e[2], ordering=ordering)
            rowkeys.sort(key=_sortkey)
        else:
            rowkeys.sort(key=p)
        ranks = rank_sorted_iterable([p(rowkey) for rowkey, _, _ in rowkeys])
        row_ranks = sorted(zip((rownum for _, rownum, _ in rowkeys), ranks), key=p)
        row_ranks = [rank for rownum, rank in row_ranks]
        c.setValues(sheet.rows, *[p(row_rank) for row_rank in row_ranks])

Sheet.addCommand('', 'addcol-rank-sheet', 'sheet.addcol_sheetrank(rows)', 'add column with the rank of each row based on its key columns')



from visidata import Sheet, anytype

@Sheet.api
def hint_type_int(sheet):
    c = sheet.cursorCol
    if c.type is anytype:
        v = c.getTypedValue(sheet.cursorRow)
        if isinstance(v, int) or int(v) is not None:
            return 2, '[:onclick type-int]Set column type to integer[/] with `#`.'

@Sheet.api
def hint_type_float(sheet):
    c = sheet.cursorCol
    if c.type is anytype:
        v = c.getTypedValue(sheet.cursorRow)
        if isinstance(v, float) or float(v) is not None:
            return 1, '[:onclick type-float]Set column type to floating point[/] with `%`.'



'ASCII theme using the first 8 colors'

from visidata import vd


vd.themes['ascii8'] = dict(
    disp_note_none='',
    disp_truncator='>',
    disp_oddspace='.',
    disp_more_left='<',
    disp_more_right='>',
    disp_error_val='',
    disp_ambig_width=1,

    disp_pending='',
    disp_note_pending=':',
    disp_note_fmtexc='?',
    disp_note_getexc='!',
    disp_note_typeexc='!',

    color_note_pending='bold magenta',
    color_note_type='yellow',
    color_note_row='yellow',

    disp_column_sep='|',
    disp_keycol_sep='|',
    disp_rowtop_sep='|',
    disp_rowmid_sep='|',
    disp_rowbot_sep='|',
    disp_rowend_sep='|',
    disp_keytop_sep='|',
    disp_keymid_sep='|',
    disp_keybot_sep='|',
    disp_endtop_sep='|',
    disp_endmid_sep='|',
    disp_endbot_sep='|',
    disp_selected_note='+',
    disp_sort_asc='^^^^^^',
    disp_sort_desc='vvvvvv',
    color_default='white on black',
    color_default_hdr='bold',
    color_bottom_hdr='underline',
    color_current_row='reverse',
    color_current_col='bold',
    color_current_hdr='bold reverse',
    color_column_sep='blue',
    color_key_col='cyan',
    color_hidden_col='8',
    color_selected_row='yellow',
    color_edit_cell='white',
    color_graph_hidden='blue',
    color_graph_selected='bold',
    color_status_replay='green',

    color_graph_axis='bold',
    color_sidebar='black on blue',
    color_add_pending='green',
    color_change_pending='reverse yellow',
    color_delete_pending='red',
    disp_rstatus_fmt=' {sheet.longname} {sheet.nRows:9d} {sheet.rowtype} {sheet.modifiedStatus} {sheet.options.disp_selected_note}{sheet.nSelectedRows}',
    disp_status_fmt='{sheet.shortcut}> {sheet.name}| ',
    disp_lstatus_max=0,
    disp_status_sep=' | ',
    color_keystrokes='bold black on cyan',
    color_status='bold black on cyan',
    color_error='red',
    color_warning='yellow',
    color_top_status='underline',
    color_active_status='black on cyan',
    color_inactive_status='8 on black',
    color_working='bold white',
    color_longname_status='black on cyan',

    color_menu='black on cyan',
    color_menu_active='yellow on black',
    color_menu_spec='black on green',
    color_menu_help='black on cyan',
    disp_menu_boxchars='||--    ||',
    disp_menu_more='>',
    disp_menu_push='+',
    disp_menu_input='_',
    disp_menu_fmt='7-bit ASCII 3-bit color',
    plot_colors = 'white',
    disp_histogram='*',
    disp_graph_reflines_x_charset='||||',
    disp_graph_reflines_y_charset='----'
)



import math

from visidata import VisiData, Canvas, Sheet, Progress, BoundingBox, Point, ColumnsSheet
from visidata import vd, asyncthread, dispwidth, colors, clipstr, ColorAttr, update_attr
from visidata.type_date import date
from statistics import median

vd.theme_option('color_graph_axis', 'bold', 'color for graph axis labels')
vd.theme_option('disp_graph_tick_x', '╵', 'character for graph x-axis ticks')
vd.theme_option('color_graph_refline', '', 'color for graph reference value lines')
vd.theme_option('disp_graph_reflines_x_charset', '▏││▕', 'charset to render vertical reference lines on graph')
vd.theme_option('disp_graph_reflines_y_charset', '▔──▁', 'charset to render horizontal reference lines on graph')
vd.theme_option('disp_graph_multiple_reflines_char', '▒', 'char to render multiple parallel reflines')


@VisiData.api
def numericCols(vd, cols):
    return [c for c in cols if vd.isNumeric(c)]


class InvertedCanvas(Canvas):
    @asyncthread
    def render_async(self):
        self.plot_elements(invert_y=True)

    def fixPoint(self, plotterPoint, canvasPoint):
        'adjust visibleBox.xymin so that canvasPoint is plotted at plotterPoint'
        self.visibleBox.xmin = canvasPoint.x - self.canvasW(plotterPoint.x-self.plotviewBox.xmin)
        self.visibleBox.ymin = canvasPoint.y - self.canvasH(self.plotviewBox.ymax-plotterPoint.y)
        self.resetBounds()

    def rowsWithin(self, plotter_bbox):
        return super().rowsWithin(plotter_bbox, invert_y=True)

    def zoomTo(self, bbox):
        super().zoomTo(bbox)
        self.fixPoint(Point(self.plotviewBox.xmin, self.plotviewBox.ymin),
                      Point(bbox.xmin, bbox.ymax))
        self.resetBounds()

    def scaleY(self, canvasY) -> int:
        'returns a plotter y coordinate for a canvas y coordinate, with the y direction inverted'
        return self.plotviewBox.ymax-round((canvasY-self.visibleBox.ymin)*self.yScaler)

    def unscaleY(self, plotterY_inverted):
        'performs the inverse of scaleY, returns a canvas y coordinate'
        return (self.plotviewBox.ymax-plotterY_inverted)/self.yScaler + self.visibleBox.ymin

    @property
    def canvasMouse(self):
        p = super().canvasMouse
        if not p: return None
        p.y = self.unscaleY(self.plotterMouse.y)
        return p

    def calcTopCursorY(self):
        'ymin for the cursor that will align its top with the top edge of the graph'
        return self.visibleBox.ymax - self.cursorBox.h

    def calcBottomCursorY(self):
        # Shift by 1 plotter pixel, like with goTopCursorY for Canvas. But shift in the
        # opposite direction, because the y-coordinate system is inverted.
        'ymin for the cursor that will align its bottom with the bottom edge of the graph'
        return self.visibleBox.ymin - (1/4 * self.canvasCharHeight)

    def startCursor(self):
        res = super().startCursor()
        if not res: return None
        # Since the y coordinates for plotting increase in the opposite
        # direction from Canvas, the cursor has to be shifted.
        self.cursorBox.ymin -= self.canvasCharHeight

# provides axis labels, legend
class GraphSheet(InvertedCanvas):
    rowtype = 'points'

    def __init__(self, *names, **kwargs):
        self.ylabel_maxw = 0
        super().__init__(*names, **kwargs)

        self.reflines_x = []
        self.reflines_y = []
        self.reflines_char_x = {}    # { x value in character coordinates -> character to use to draw that vertical line }
        self.reflines_char_y = {}    # { y value in character coordinates -> character to use to draw that horizontal line }

        vd.numericCols(self.xcols) or vd.fail('at least one numeric key col necessary for x-axis')
        self.ycols or vd.fail('%s is non-numeric' % '/'.join(yc.name for yc in kwargs.get('ycols')))

    def resetCanvasDimensions(self, windowHeight, windowWidth):
        if self.left_margin < self.ylabel_maxw:
            self.left_margin = self.ylabel_maxw
        super().resetCanvasDimensions(windowHeight, windowWidth)

    @asyncthread
    def reload(self):
        nerrors = 0
        nplotted = 0

        self.reset()
        self.row_order = {}

        vd.status('loading data points')
        catcols = [c for c in self.xcols if not vd.isNumeric(c)]
        numcols = vd.numericCols(self.xcols)
        for ycol in self.ycols:
            for rownum, row in enumerate(Progress(self.sourceRows, 'plotting')):  # rows being plotted from source
                try:
                    k = tuple(c.getValue(row) for c in catcols) if catcols else (ycol.name,)

                    # convert deliberately to float (to e.g. linearize date)
                    graph_x = float(numcols[0].type(numcols[0].getValue(row))) if numcols else rownum
                    graph_y = ycol.type(ycol.getValue(row))

                    attr = self.plotColor(k)
                    self.point(graph_x, graph_y, attr, row)
                    self.row_order[self.source.rowid(row)] = rownum
                    nplotted += 1
                except Exception as e:
                    nerrors += 1
                    if vd.options.debug:
                        vd.exceptionCaught(e)


        vd.status('loaded %d points (%d errors)' % (nplotted, nerrors))

        self.xzoomlevel=self.yzoomlevel=1.0
        self.resetBounds()

    def draw(self, scr):
        windowHeight, windowWidth = scr.getmaxyx()
        if self.needsRefresh:
            self.render(windowHeight, windowWidth)

        # required because we use clear_empty_squares=False for draw_pixels()
        self.draw_empty(scr)
        # draw reflines first so pixels draw over them
        self.draw_reflines(scr)
        # use clear_empty_squares to keep reflines
        self.draw_pixels(scr, clear_empty_squares=False)
        self.draw_labels(scr)

    def draw_reflines(self, scr):
        cursorBBox = self.plotterCursorBox
        # draws only on character cells that have reflines, leaves other cells unaffected
        for char_y in range(0, self.plotheight//4):
            has_y_line = char_y in self.reflines_char_y.keys()
            for char_x in range(0, self.plotwidth//2):
                has_x_line = char_x in self.reflines_char_x.keys()
                if has_x_line or has_y_line:
                    cattr = colors.color_graph_refline
                    if has_x_line:
                        ch = self.reflines_char_x[char_x]
                        # where two lines cross, draw the vertical line, not the horizontal one
                    elif has_y_line:
                        ch = self.reflines_char_y[char_y]
                    # draw cursor
                    if cursorBBox.contains(char_x*2, char_y*4) or \
                        cursorBBox.contains(char_x*2+1, char_y*4+3):
                        cattr = update_attr(cattr, colors.color_current_row)
                    scr.addstr(char_y, char_x, ch, cattr.attr)

    def resetBounds(self, refresh=True):
        super().resetBounds(refresh=False)
        self.createLabels()
        if refresh:
            self.refresh()

    def moveToRow(self, rowstr):
        ymin, ymax = map(float, map(self.parseY, rowstr.split()))
        self.cursorBox.ymin = ymin
        self.cursorBox.h = ymax-ymin
        return True

    def plot_elements(self, invert_y=True):
        self.plot_reflines()
        super().plot_elements(invert_y=True)

    def plot_reflines(self):
        self.reflines_char_x = {}
        self.reflines_char_y = {}

        bb = self.visibleBox
        xmin, ymin, xmax, ymax = bb.xmin, bb.ymin, bb.xmax, bb.ymax

        for data_y in self.reflines_y:
            data_y = float(data_y)
            if data_y >= ymin and data_y <= ymax:
                char_y, offset = divmod(self.scaleY(data_y), 4)
                chars = self.options.disp_graph_reflines_y_charset
                # if we're drawing two different reflines in the same square, fill it with a different char
                if char_y in self.reflines_char_y and self.reflines_char_y[char_y] != chars[offset]:
                    self.reflines_char_y[char_y] = vd.options.disp_graph_multiple_reflines_char
                else:
                    self.reflines_char_y[char_y] = chars[offset]

        for data_x in self.reflines_x:
            data_x = float(data_x)
            if data_x >= xmin and data_x <= xmax:
                plot_x = self.scaleX(data_x)
                # plot_x is an integer count of plotter pixels, and each character box has 2 plotter pixels
                char_x = plot_x // 2
                # To subdivide the 2 plotter pixels per square into 4 zones, we have to first multiply by 2.
                offset = 2*plot_x % 4
                chars = self.options.disp_graph_reflines_x_charset
                # if we're drawing two different reflines in the same square, fill it with a different char
                if char_x in self.reflines_char_x and self.reflines_char_x[char_x] != chars[offset]:
                    self.reflines_char_y[char_x] = vd.options.disp_graph_multiple_reflines_char
                else:
                    self.reflines_char_x[char_x] = chars[offset]

    def moveToCol(self, colstr):
        xmin, xmax = map(float, map(self.parseX, colstr.split()))
        self.cursorBox.xmin = xmin
        self.cursorBox.w = xmax-xmin
        return True

    def formatX(self, amt):
        return ','.join(xcol.format(xcol.type(amt)) for xcol in self.xcols if vd.isNumeric(xcol))

    def formatY(self, amt):
        srccol = self.ycols[0]
        return srccol.format(srccol.type(amt))

    def formatXLabel(self, amt):
        if self.xzoomlevel < 1:
            labels = []
            for xcol in self.xcols:
                if vd.isNumeric(xcol):
                    col_amt = float(amt) if xcol.type is int else xcol.type(amt)
                else:
                    continue
                labels.append(xcol.format(col_amt))
            return ','.join(labels)
        else:
            return self.formatX(amt)

    def formatYLabel(self, amt):
        srccol = self.ycols[0]
        if srccol.type is int and self.yzoomlevel < 1:
            return srccol.format(float(amt))
        else:
            return self.formatY(amt)

    def parseX(self, txt):
        return self.xcols[0].type(txt)

    def parseY(self, txt):
        return self.ycols[0].type(txt)

    def add_y_axis_label(self, frac):
        label_data_y = self.visibleBox.ymin + frac*self.visibleBox.h
        txt = self.formatYLabel(label_data_y)
        w = (dispwidth(txt)+1)*2
        if self.ylabel_maxw < w:
            self.ylabel_maxw = w
        y = self.scaleY(label_data_y)

        # plot y-axis labels on the far left of the canvas, but within the plotview height-wise
        self.plotlabel(0, y, txt, 'graph_axis')

    def add_x_axis_label(self, frac):
        label_data_x = self.visibleBox.xmin + frac*self.visibleBox.w
        txt = self.formatXLabel(label_data_x)
        tick = vd.options.disp_graph_tick_x or ''

        # plot x-axis labels below the plotviewBox.ymax, but within the plotview width-wise
        x = self.scaleX(label_data_x)

        if frac < 1.0:
            txt = tick + txt
        else:
            right_margin = self.plotwidth - 1 - self.plotviewBox.xmax
            if (dispwidth(txt)+dispwidth(tick))*2 <= right_margin:
                txt = tick + txt
            else:
                # shift rightmost label to be left of its tick
                x -= dispwidth(txt)*2
                if len(tick) == 0:
                    x += 1
                txt = txt + tick

        self.plotlabel(x, self.plotviewBox.ymax+4, txt, 'graph_axis')

    def createLabels(self):
        self.gridlabels = []
        self.ylabel_maxw = self.leftMarginPixels

        # y-axis
        self.add_y_axis_label(1.00)
        self.add_y_axis_label(0.75)
        self.add_y_axis_label(0.50)
        self.add_y_axis_label(0.25)
        self.add_y_axis_label(0.00)

        # x-axis
        self.add_x_axis_label(1.00)
        self.add_x_axis_label(0.75)
        self.add_x_axis_label(0.50)
        self.add_x_axis_label(0.25)
        self.add_x_axis_label(0.00)

        # TODO: if 0 line is within visible bounds, explicitly draw the axis
        # TODO: grid lines corresponding to axis labels

        xname = ','.join(xcol.name for xcol in self.xcols if vd.isNumeric(xcol)) or 'row#'
        xname, _ = clipstr(xname, self.left_margin//2-2)
        self.plotlabel(0, self.plotviewBox.ymax+4, xname+'»', 'graph_axis')

    def rowsWithin(self, plotter_bbox):
        'return list of deduped rows within plotter_bbox'
        rows = super().rowsWithin(plotter_bbox)
        return sorted(rows, key=lambda r: self.row_order[self.source.rowid(r)])

    def draw_refline_x(self):
        xcol = vd.numericCols(self.xcols)[0]
        xtype = xcol.type
        val = median(xcol.getValues(self.sourceRows))
        suggested = format_input_value(val, xtype)
        xstrs = vd.input("add line(s) at x = ", type="reflinex", value=suggested, defaultLast=True).split()

        for xstr in xstrs:
            vals = [ v.strip() for v in xstr.split(',') ]
            if len(vals) != len(self.xcols):
                vd.fail(f'must have {len(self.xcols)} x values, had {len(vals)} values: {xstr}')
            self.reflines_x += [xtype(val) for xcol, val in zip(self.xcols, vals) if xtype(val) not in self.reflines_x ]
        self.refresh()

    def draw_refline_y(self):
        ytype = self.ycols[0].type
        val = median(self.ycols[0].getValues(self.sourceRows))
        suggested = format_input_value(val, ytype)
        ystrs = vd.input("add line(s) at y = ", type="refliney", value=suggested, defaultLast=True).split()

        self.reflines_y += [ ytype(y) for y in ystrs if ytype(y) not in self.reflines_y ]
        self.refresh()

    def erase_refline_x(self):
        if len(self.reflines_x) == 0:
            vd.fail(f'no x refline to erase')
        xtype = vd.numericCols(self.xcols)[0].type
        suggested = format_input_value(self.reflines_x[0], xtype)

        xstrs = vd.input('remove line(s) at x = ', value=suggested, type='reflinex', defaultLast=True).split()
        for input_x in xstrs:
            self.reflines_x.remove(xtype(input_x))
        self.refresh()

    def erase_refline_y(self):
        ytype = self.ycols[0].type
        suggested = format_input_value(self.reflines_y[0], ytype) if self.reflines_y else ''
        ystrs = vd.input('remove line(s) at y = ', value=suggested, type='refliney', defaultLast=True).split()
        for y in ystrs:
            try:
                self.reflines_y.remove(ytype(y))
            except ValueError:
                vd.fail(f'value {y} not in reflines_y')
        self.refresh()

def format_input_value(val, type):
    '''format a value for entry into vd.input(), so its representation has no spaces and no commas'''
    if type is date:
        return val.strftime('%Y-%m-%d')
    else:
        return str(val)


Sheet.addCommand('.', 'plot-column', 'vd.push(GraphSheet(sheet.name, "graph", source=sheet, sourceRows=rows, xcols=keyCols, ycols=numericCols([cursorCol])))', 'plot current numeric column vs key columns; numeric key column is used for x-axis, while categorical key columns determine color')
Sheet.addCommand('g.', 'plot-numerics', 'vd.push(GraphSheet(sheet.name, "graph", source=sheet, sourceRows=rows, xcols=keyCols, ycols=numericCols(nonKeyVisibleCols)))', 'plot a graph of all visible numeric columns vs key columns')
ColumnsSheet.addCommand('g.', 'plot-source-selected', 'vd.push(GraphSheet(sheet.source[0].name, "graph", source=source[0], sourceRows=source[0].rows, xcols=source[0].keyCols, ycols=numericCols(selectedRows)))', 'plot a graph of all selected columns vs key columns on source sheet')

# swap directions of up/down
InvertedCanvas.addCommand(None, 'go-up', 'if cursorBox: sheet.cursorBox.ymin += cursorBox.h', 'move cursor up by its height')
InvertedCanvas.addCommand(None, 'go-down', 'if cursorBox: sheet.cursorBox.ymin -= cursorBox.h', 'move cursor down by its height')
InvertedCanvas.addCommand(None, 'go-top',  'if cursorBox: sheet.cursorBox.ymin = sheet.calcTopCursorY()', 'move cursor to top edge of visible canvas')
InvertedCanvas.addCommand(None, 'go-bottom', 'if cursorBox: sheet.cursorBox.ymin = sheet.calcBottomCursorY()', 'move cursor to bottom edge of visible canvas')
InvertedCanvas.addCommand(None, 'go-pagedown', 't=(visibleBox.ymax-visibleBox.ymin); sheet.cursorBox.ymin -= t; sheet.visibleBox.ymin -= t; sheet.refresh()', 'move cursor down to next visible page')
InvertedCanvas.addCommand(None, 'go-pageup', 't=(visibleBox.ymax-visibleBox.ymin); sheet.cursorBox.ymin += t; sheet.visibleBox.ymin += t; sheet.refresh()', 'move cursor up to previous visible page')

InvertedCanvas.addCommand(None, 'go-down-small', 'sheet.cursorBox.ymin -= canvasCharHeight', 'move cursor down one character')
InvertedCanvas.addCommand(None, 'go-up-small', 'sheet.cursorBox.ymin += canvasCharHeight', 'move cursor up one character')

InvertedCanvas.addCommand(None, 'resize-cursor-shorter', 'sheet.cursorBox.h -= canvasCharHeight', 'decrease cursor height by one character')
InvertedCanvas.addCommand(None, 'resize-cursor-taller', 'sheet.cursorBox.h += canvasCharHeight', 'increase cursor height by one character')


@GraphSheet.api
def set_y(sheet, s):
    ymin, ymax = map(float, map(sheet.parseY, s.split()))
    sheet.zoomTo(BoundingBox(sheet.visibleBox.xmin, ymin, sheet.visibleBox.xmax, ymax))
    sheet.refresh()

@GraphSheet.api
def set_x(sheet, s):
    xmin, xmax = map(float, map(sheet.parseX, s.split()))
    sheet.zoomTo(BoundingBox(xmin, sheet.visibleBox.ymin, xmax, sheet.visibleBox.ymax))
    sheet.refresh()

Canvas.addCommand('y', 'resize-y-input', 'sheet.set_y(input("set ymin ymax="))', 'set ymin/ymax on graph axes')
Canvas.addCommand('x', 'resize-x-input', 'sheet.set_x(input("set xmin xmax="))', 'set xmin/xmax on graph axes')

GraphSheet.addCommand('gx', 'draw-refline-x', 'sheet.draw_refline_x()', 'draw a vertical line at x-values (space-separated)')
GraphSheet.addCommand('gy', 'draw-refline-y', 'sheet.draw_refline_y()', 'draw a horizontal line at y-values (space-separated)')
GraphSheet.addCommand('zx', 'erase-refline-x', 'sheet.erase_refline_x()', 'remove a horizontal line at x-values (space-separated)')
GraphSheet.addCommand('zy', 'erase-refline-y', 'sheet.erase_refline_y()', 'remove a vertical line at y-values (space-separated)')
GraphSheet.addCommand('gzx', 'erase-reflines-x', 'sheet.reflines_x = []; sheet.refresh()', 'erase all vertical x-value lines')
GraphSheet.addCommand('gzy', 'erase-reflines-y', 'sheet.reflines_y = []; sheet.refresh()', 'erase any horizontal y-value lines')

vd.addGlobals({
    'GraphSheet': GraphSheet,
    'InvertedCanvas': InvertedCanvas,
})

vd.addMenuItems('''
    Plot > Graph > current column > plot-column
    Plot > Graph > all numeric columns > plot-numerics
    Plot > Refline > draw at x coord (vert) > draw-refline-x
    Plot > Refline > draw at y coord (horiz) > draw-refline-y
    Plot > Refline > erase at x coord (vert) > erase-refline-x
    Plot > Refline > erase at y coord (horiz) > erase-refline-y
    Plot > Refline > erase all x (vert) > erase-reflines-x
    Plot > Refline > erase all y (horiz) > erase-reflines-y
''')



from visidata import VisiData, vd, Sheet, Path, Column, ItemColumn, BaseSheet, anytype
from itertools import chain

@VisiData.api
def open_h5(vd, p):
    return Hdf5ObjSheet(p.base_stem, source=p)

VisiData.open_hdf5 = VisiData.open_h5

vd.option('hdf5_matrix_enumerate', False, 'enumerate matrix rows and columns')

class Hdf5ObjSheet(Sheet):
    'Support sheets in HDF5 format.'

    def iterload(self):
        h5py = vd.importExternal('h5py')
        source = self.source
        if isinstance(self.source, Path):
            source = h5py.File(str(self.source), 'r')

        self.columns = []
        if isinstance(source, h5py.Group):
            self.rowtype = 'sheets'
            self.columns = [
                Column(source.name, type=str, getter=lambda col,row: row.source.name.split('/')[-1], keycol=1),
                Column('type', type=str, getter=lambda col,row: type(row.source).__name__),
                Column('nItems', type=int, getter=lambda col,row: len(row.source)),
            ]
            self.recalc()
            for k, v in source.items():
                yield Hdf5ObjSheet(self.name, k, source=v)
        elif isinstance(source, h5py.Dataset):
            if len(source.shape)==1:
                if source.dtype.names:
                    for i, (colname, fmt, *_) in enumerate(source.dtype.descr):
                        if not colname:
                            colname = f"col{i}"
                        ctype = _guess_type(fmt)
                        self.addColumn(ItemColumn(colname, i, type=ctype))
                    yield from source  # copy
                else:
                    self.addColumn(ItemColumn(source.name, 0))
                    for v in source:
                        yield [v]
            elif len(source.shape)==2:
                matrix_enumerate = bool(self.options.hdf5_matrix_enumerate)

                ncols = source.shape[1]
                ctype = _guess_type(source.dtype.descr[0][1])

                if matrix_enumerate:
                    self.addColumn(ItemColumn("row", 0, width=8, keycol=1, type=int), index=0)
                    for i in range(ncols):
                        self.addColumn(ItemColumn(f'col{i}', i+1, width=8, type=ctype), index=i+1)
                    self.recalc()
                    yield from list(list((chain((i,), row))) for i, row in enumerate(source))
                else:
                    for i in range(ncols):
                        self.addColumn(ItemColumn('', i, width=8, type=ctype), index=i)
                    self.recalc()
                    yield from source  # copy
            else:
                vd.fail('too many dimensions in shape %s' % str(source.shape))
        else:
            vd.fail(f"too many dimensions in shape {source.shape}")


    def openRow(self, row):
        h5py = vd.importExternal('h5py')
        if isinstance(row, BaseSheet):
            return row
        if isinstance(row, h5py.HLObject):
            return Hdf5ObjSheet(row)

        numpy = vd.importExternal('numpy')
        from .npy import NpySheet
        if isinstance(row, numpy.ndarray):
            return NpySheet(None, npy=row)

def _guess_type(fmt):
    if 'i' in fmt or 'u' in fmt:
        return int
    elif 'f' in fmt:
        return float
    return anytype

Hdf5ObjSheet.addCommand('A', 'dive-metadata', 'vd.push(SheetDict(cursorRow.name + "_attrs", source=cursorRow.attrs))', 'open metadata sheet for object referenced in current row')



#!/usr/bin/env python3

from setuptools import setup, find_packages

# Note: use `python3 visidata/apps/setup.py install` from the root directory

__version__ = '0.2-dev'

setup(name='vgit',
      version=__version__,
      description='a sleek terminal user interface for git',
      # long_description=open('README.md').read(),
      install_requires=['sh<2'], # visidata
      packages=find_packages(exclude=["tests"]),
      scripts=['vgit'],
      entry_points={'visidata.plugins': 'vgit=visidata.apps.vgit'},
      author='Saul Pwanson',
      author_email='vgit@saul.pw',
      url='https://github.com/saulpw/visidata/vgit',
      license='GPLv3',
      python_requires='>=3.7',
      classifiers=[
          'Development Status :: 2 - Pre-Alpha',
          'Environment :: Console',
          'Environment :: Console :: Curses',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'Intended Audience :: Information Technology',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3',
          'Topic :: Utilities',
          'Topic :: Software Development :: Version Control',
          'Topic :: Terminals'
      ],
      keywords=('console textpunk git version-control curses visidata tui terminal'),
      )
